#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli

"""
This package contains analyic functions useful for Zernike polynomials and
the aperture distribution function.
"""

# Shortcuts for the two Zernike polynomials functions
from .zernike import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This sub-package contains all necessary functions to compute the aperture
distribution and the (field) radiation pattern, and is ready to be used by the
least squares minimization.
"""

from .aperture import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import numpy as np

__all__ = [
    'cart2pol', 'wavevector2degrees', 'wavevector2radians', 'co_matrices',
    'line_equation', 'rms'
    ]


def cart2pol(x, y):
    """
    Transformation from Cartesian to polar coordinates, in two dimensions.

    Parameters
    ----------
    x : `~numpy.ndarray`
        Grid value for the :math:`x` variable.
    y : `~numpy.ndarray`
        Grid value for the :math:`y` variable.

    Returns
    -------
    rho : `~numpy.ndarray`
        Grid value for the radial variable.
    theta : `~numpy.ndarray`
        Grid value for the angular variable, in radians.
    """

    rho = np.sqrt(x ** 2 + y ** 2)
    theta = np.arctan2(y, x)

    return rho, theta


def wavevector2degrees(u, wavel):
    """
    Transformation from a wave-vector 1 / m units to degrees.

    Parameters
    ----------
    u : `~numpy.ndarray`
        Wave-vector, result from FFT2 in 1 / m units.
    wavel : `~numpy.ndarray`
        Wavelength in meters.

    Returns
    -------
    wavevector_degrees : `~numpy.ndarray`
        Wave-vector in degrees.
    """

    wavevector_degrees = np.degrees(u * wavel)

    return wavevector_degrees


def wavevector2radians(u, wavel):
    """
    Transformation from a wave-vector 1 / m units to radians.

    Parameters
    ----------
    u : `~numpy.ndarray`
        Wave-vector, result from FFT2 in 1 / m units.
    wavel : `~numpy.ndarray`
        Wavelength in meters.

    Returns
    -------
    wavevector_degrees : `~numpy.ndarray`
        Wave-vector in radians.
    """

    wavevector_radians = wavevector2degrees(u, wavel) * np.pi / 180

    return wavevector_radians


def co_matrices(res, jac, n_pars):
    """
    Computes the estimated Variance-Covariance and the Correlation matrices.
    Assuming an estimator normally distributed (and consistent).

    Parameters
    ----------
    res : `~numpy.ndarray`
        Last residual evaluation from a fit procedure (least squares
        minimization), residual understood as data - model.
    jac : `~numpy.ndarray`
        Last Jacobian matrix evaluation from a fit procedure.
    n_pars : `int`
        Total number of parameters in the model (only the ones that have been
        fitted). It is related to the degrees of freedom.

    Returns
    -------
    cov : `~numpy.ndarray`
        Variance-Covariance matrix. Dimensions :math:`n \\times p`.
    corr : `~numpy.ndarray`
        Correlation matrix. Dimensions :math:`n \\times p`.
    """

    m = res.size  # Number of data points used in the fit
    p = m - n_pars  # Degrees of freedom

    # Variance-Covariance matrix
    cov = np.dot(res.T, res) / p * np.linalg.inv(np.dot(jac.T, jac))

    sigmas2 = np.diag(np.diag(cov))  # Estimated error variance (sigma ** 2)
    D = np.linalg.inv(np.sqrt(sigmas2))  # Inverted diagonal variance matrix
    corr = np.dot(np.dot(D, cov), D)  # Correlation matrix

    return cov, corr


def line_equation(P1, P2, x):
    """
    Computes the linear equation solution for the :math:`y` vector values
    given two data points, :math:`P_1 = (x_1, y_1)` and :math:`P_2 = (x_2,
    y_2)`, and  the :math:`x` vector.

    Parameters
    ----------
    P1 : `tuple`
        First point that belongs to the desired linear equation.
    P2 : `tuple`
        Second point that belongs to the desired linear equation.
    x : `~numpy.ndarray`
        Data points from the :math:`x`-axis.

    Returns
    -------
    y  : `~numpy.ndarray`
        Linear equation :math:`y`-data points.
    """

    (x1, y1) = P1
    (x2, y2) = P2

    y = (y2 - y1) / (x2 - x1) * (x - x1) + y1

    return y


def rms(phase):
    """
    Computes the root-mean-square value from a aperture phase distribution
    map, :math:`\\varphi(x, y)`.

    Parameters
    ----------
    phase : `~numpy.ndarray`
        One or two dimensional array for the aperture phase distribution.
    """

    # To remove elements out limit radius
    nonzero_values = phase[np.nonzero(phase)]

    return np.sqrt(np.sum(np.square(nonzero_values)) / nonzero_values.size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This sub-package contains the geometrical properties of radio telescopes. There
are ready to use specific telescope structures and a
`~pyoof.telgeometry.block_manual` for different set ups.
"""

from .telgeometry import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy import interpolate
from astropy.io import ascii
from astropy.utils.data import get_pkg_data_filename
import os
import yaml
from .aperture import radiation_pattern, phase
from .math_functions import wavevector2degrees, wavevector2radians
from .aux_functions import uv_ratio

__all__ = [
    'plot_beam', 'plot_data', 'plot_phase', 'plot_variance', 'plot_fit_path'
    ]

# Plot style added from relative path
plt.style.use(get_pkg_data_filename('data/pyoof.mplstyle'))


def plot_beam(
    params, d_z, wavel, illum_func, telgeo, resolution, box_factor, plim_rad,
    angle, title
        ):
    """
    Beam maps, :math:`P_\\mathrm{norm}(u, v)`, figure given fixed
    ``I_coeff`` coefficients and ``K_coeff`` set of coefficients. It is the
    straight forward result from a least squares minimization
    (`~pyoof.fit_beam`). There will be three maps, for three radial offsets,
    :math:`d_z^-`, :math:`0` and :math:`d_z^+` (in meters).

    Parameters
    ----------
    params : `~numpy.ndarray`
        Two stacked arrays, the illumination and Zernike circle polynomials
        coefficients. ``params = np.hstack([I_coeff, K_coeff])``.
    d_z : `list`
        Radial offset :math:`d_z`, added to the sub-reflector in meters. This
        characteristic measurement adds the classical interference pattern to
        the beam maps, normalized squared (field) radiation pattern, which is
        an out-of-focus property. The radial offset list must be as follows,
        ``d_z = [d_z-, 0., d_z+]`` all of them in meters.
    wavel : `float`
        Wavelength, :math:`\\lambda`, of the observation in meters.
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with the key **I_coeff**. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.
    telgeo : `list`
        List that contains the blockage distribution, optical path difference
        (OPD) function, and the primary radius (`float`) in meters. The list
        must have the following order, ``telego = [block_dist, opd_func, pr]``.
    resolution : `int`
        Fast Fourier Transform resolution for a rectangular grid. The input
        value has to be greater or equal to the telescope resolution and with
        power of 2 for faster FFT processing. It is recommended a value higher
        than ``resolution = 2 ** 8``.
    box_factor : `int`
        Related to the FFT resolution (**resolution** key), defines the image
        pixel size level. It depends on the primary radius, ``pr``, of the
        telescope, e.g. a ``box_factor = 5`` returns ``x = np.linspace(-5 *
        pr, 5 * pr, resolution)``, an array to be used in the FFT2
        (`~numpy.fft.fft2`).
    plim_rad : `~numpy.ndarray`
        Contains the maximum values for the :math:`u` and :math:`v`
        wave-vectors, it can be in degrees or radians depending which one is
        chosen in **angle** key. The `~numpy.ndarray` must be in the following
        order, ``plim_rad = np.array([umin, umax, vmin, vmax])``.
    angle : `str`
        Angle unit, it can be ``'degrees'`` or ``'radians'``.
    title : `str`
        Figure title.

    Returns
    -------
    fig : `~matplotlib.figure.Figure`
        The three beam maps plotted from the input parameters. Each map with a
        different offset :math:`d_z` value. From left to right, :math:`d_z^-`,
        :math:`0` and :math:`d_z^+`.
    """

    I_coeff = params[:4]
    K_coeff = params[4:]

    u, v, F = [], [], []
    for _d_z in d_z:

        _u, _v, _F = radiation_pattern(
            K_coeff=K_coeff,
            I_coeff=I_coeff,
            d_z=_d_z,
            wavel=wavel,
            illum_func=illum_func,
            telgeo=telgeo,
            resolution=resolution,
            box_factor=box_factor
            )

        u.append(_u)
        v.append(_v)
        F.append(_F)

    u = np.array(u)
    v = np.array(v)
    power_pattern = np.abs(F) ** 2
    power_norm = np.array(
        [power_pattern[i] / power_pattern[i].max() for i in range(3)]
        )

    # Limits, they need to be transformed to degrees
    if plim_rad is None:
        pr = telgeo[2]  # primary reflector radius
        bw = 1.22 * wavel / (2 * pr)  # Beamwidth radians
        size_in_bw = bw * 8

        # Finding central point for shifted maps
        uu, vv = np.meshgrid(_u, _v)
        u_offset = uu[power_norm[1] == power_norm[1].max()][0]
        v_offset = vv[power_norm[1] == power_norm[1].max()][0]

        u_offset = wavevector2radians(u_offset, wavel)
        v_offset = wavevector2radians(v_offset, wavel)

        plim_rad = [
            -size_in_bw + u_offset, size_in_bw + u_offset,
            -size_in_bw + v_offset, size_in_bw + v_offset
            ]

    if angle == 'degrees':
        plim_angle = np.degrees(plim_rad)
        u_angle = wavevector2degrees(u, wavel)
        v_angle = wavevector2degrees(v, wavel)
    if angle == 'radians':
        plim_angle = plim_rad
        u_angle = wavevector2radians(u, wavel)
        v_angle = wavevector2radians(v, wavel)

    plim_u, plim_v = plim_angle[:2], plim_angle[2:]

    subtitle = [
        '$P_{\\textrm{\\scriptsize{norm}}}(u,v)$ $d_z=' +
        str(round(d_z[i], 3)) + '$ m' for i in range(3)
        ]

    fig, ax = plt.subplots(ncols=3, figsize=uv_ratio(plim_u, plim_v))

    for i in range(3):

        extent = [
            u_angle[i].min(), u_angle[i].max(),
            v_angle[i].min(), v_angle[i].max()
            ]

        im = ax[i].imshow(power_norm[i], extent=extent, vmin=0, vmax=1)
        ax[i].contour(u_angle[i], v_angle[i], power_norm[i], 10)

        divider = make_axes_locatable(ax[i])
        cax = divider.append_axes("right", size="3%", pad=0.03)
        cb = fig.colorbar(im, cax=cax)
        cb.formatter.set_powerlimits((0, 0))
        cb.ax.yaxis.set_offset_position('left')
        cb.update_ticks()

        ax[i].set_title(subtitle[i])
        ax[i].set_ylabel('$v$ ' + angle)
        ax[i].set_xlabel('$u$ ' + angle)
        ax[i].set_ylim(*plim_v)
        ax[i].set_xlim(*plim_u)
        ax[i].grid('off')

    fig.suptitle(title)
    fig.tight_layout()

    return fig


def plot_data(u_data, v_data, beam_data, d_z, angle, title, res_mode):
    """
    Real data beam maps, :math:`P^\\mathrm{obs}(x, y)`, figures given
    given 3 out-of-focus radial offsets, :math:`d_z`.

    Parameters
    ----------
    u_data : `~numpy.ndarray`
        :math:`x` axis value for the 3 beam maps in radians. The values have
        to be flatten, in one dimension, and stacked in the same order as the
        ``d_z = [d_z-, 0., d_z+]`` values from each beam map.
    v_data : `~numpy.ndarray`
        :math:`y` axis value for the 3 beam maps in radians. The values have
        to be flatten, one dimensional, and stacked in the same order as the
        ``d_z = [d_z-, 0., d_z+]`` values from each beam map.
    beam_data : `~numpy.ndarray`
        Amplitude value for the beam map in mJy. The values have to be
        flatten, one dimensional, and stacked in the same order as the ``d_z =
        [d_z-, 0., d_z+]`` values from each beam map. If ``res_mode = False``,
        the beam map will be normalized.
    d_z : `list`
        Radial offset :math:`d_z`, added to the sub-reflector in meters. This
        characteristic measurement adds the classical interference pattern to
        the beam maps, normalized squared (field) radiation pattern, which is
        an out-of-focus property. The radial offset list must be as follows,
        ``d_z = [d_z-, 0., d_z+]`` all of them in meters.
    wavel : `float`
        Wavelength, :math:`\\lambda`, of the observation in meters.
    angle : `str`
        Angle unit, it can be ``'degrees'`` or ``'radians'``.
    title : `str`
        Figure title.
    res_mode : `bool`
        If ``True`` the beam map will not be normalized. This feature is used
        to compare the residual outputs from the least squares minimization
        (`~pyoof.fit_beam`).

    Returns
    -------
    fig : `~matplotlib.figure.Figure`
        Figure from the three observed beam maps. Each map with a different
        offset :math:`d_z` value. From left to right, :math:`d_z^-`, :math:`0`
        and :math:`d_z^+`.
    """

    if not res_mode:
        # Power pattern normalization
        beam_data = [beam_data[i] / beam_data[i].max() for i in range(3)]
    # input u and v are in radians
    uv_title = angle

    if angle == 'degrees':
        u_data, v_data = np.degrees(u_data), np.degrees(v_data)

    vmin = np.min(beam_data)
    vmax = np.max(beam_data)

    subtitle = [
        '$P_{\\textrm{\\scriptsize{norm}}}^{\\textrm{\\scriptsize{obs}}}' +
        '(u,v)$ $d_z={}$ m'.format(round(d_z[i], 3)) for i in range(3)
        ]

    fig, ax = plt.subplots(ncols=3, figsize=uv_ratio(u_data[0], v_data[0]))

    for i in range(3):
        # new grid for beam_data
        u_ng = np.linspace(u_data[i].min(), u_data[i].max(), 300)
        v_ng = np.linspace(v_data[i].min(), v_data[i].max(), 300)

        beam_ng = interpolate.griddata(
            # coordinates of grid points to interpolate from.
            points=(u_data[i], v_data[i]),
            values=beam_data[i],
            # coordinates of grid points to interpolate to.
            xi=tuple(np.meshgrid(u_ng, v_ng)),
            method='cubic'
            )

        extent = [u_ng.min(), u_ng.max(), v_ng.min(), v_ng.max()]
        im = ax[i].imshow(beam_ng, extent=extent, vmin=vmin, vmax=vmax)
        ax[i].contour(u_ng, v_ng, beam_ng, 10)

        divider = make_axes_locatable(ax[i])
        cax = divider.append_axes("right", size="3%", pad=0.03)

        cb = fig.colorbar(im, cax=cax)
        cb.formatter.set_powerlimits((0, 0))
        cb.ax.yaxis.set_offset_position('left')
        cb.update_ticks()

        ax[i].set_ylabel('$v$ ' + uv_title)
        ax[i].set_xlabel('$u$ ' + uv_title)
        ax[i].set_title(subtitle[i])
        ax[i].grid('off')

    fig.suptitle(title)
    fig.tight_layout()

    return fig


def plot_phase(K_coeff, notilt, pr, title):
    """
    Aperture phase distribution (phase error), :math:`\\varphi(x, y)`, figure,
    given the Zernike circle polynomial coefficients, ``K_coeff``, solution
    from the least squares minimization.

    Parameters
    ----------
    K_coeff : `~numpy.ndarray`
        Constants coefficients, :math:`K_{n\\ell}`, for each of them there is
        only one Zernike circle polynomial, :math:`U^\ell_n(\\varrho,
        \\varphi)`. The coefficients are between :math:`[-2, 2]`.
    notilt : `bool`
        Boolean to include or exclude the tilt coefficients in the aperture
        phase distribution. The Zernike circle polynomials are related to tilt
        through :math:`U^{-1}_1(\\varrho, \\varphi)` and
        :math:`U^1_1(\\varrho, \\varphi)`.
    pr : `float`
        Primary reflector radius in meters.
    title : `str`
        Figure title.

    Returns
    -------
    fig : `~matplotlib.figure.Figure`
        Aperture phase distribution parametrized in terms of the Zernike
        circle polynomials, and represented for the telescope's primary dish.
    """

    if notilt:
        cbartitle = (
            '$\\varphi_{\\scriptsize{\\textrm{no-tilt}}}(x,y)$ amplitude rad'
            )
    else:
        cbartitle = '$\\varphi(x, y)$ amplitude rad'

    extent = [-pr, pr, -pr, pr]
    _x, _y, _phase = phase(K_coeff=K_coeff, notilt=notilt, pr=pr)

    fig, ax = plt.subplots(figsize=(6, 5.8))

    levels = np.linspace(-2, 2, 9)

    im = ax.imshow(_phase, extent=extent)
    ax.contour(_x, _y, _phase, levels=levels, colors='k', alpha=0.3)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="3%", pad=0.03)
    cb = fig.colorbar(im, cax=cax)
    cb.ax.set_ylabel(cbartitle)

    ax.set_title(title)
    ax.set_ylabel('$y$ m')
    ax.set_xlabel('$x$ m')
    ax.grid('off')

    fig.tight_layout()

    return fig


def plot_variance(matrix, order, diag, cbtitle, title):
    """
    Variance-Covariance matrix or Correlation matrix figure. It returns
    the triangle figure with a color amplitude value for each element. Used to
    check/compare the correlation between the fitted parameters in a least
    squares minimization.

    Parameters
    ----------
    matrix : `~numpy.ndarray`
        Two dimensional array containing the Variance-Covariance or
        Correlation function. Output from the fit procedure.
    order : `int`
        Order used for the Zernike circle polynomial, :math:`n`.
    diag : `bool`
        If `True` it will plot the matrix diagonal.
    cbtitle : `str`
        Color bar title.
    title : `str`
        Figure title.

    Returns
    -------
    fig : `~matplotlib.figure.Figure`
        Triangle figure representing Variance-Covariance or Correlation matrix.
    """
    n = order
    N_K_coeff = (n + 1) * (n + 2) // 2
    ln = [(j, i) for i in range(0, n + 1) for j in range(-i, i + 1, 2)]
    L = np.array(ln)[:, 0]
    N = np.array(ln)[:, 1]

    params_names = [
        '$A_{E_\mathrm{a}}$', '$\mathrm{taper}_\mathrm{dB}$', '$x_0$', '$y_0$'
        ]
    for i in range(N_K_coeff):
        params_names.append('$K_{' + str(N[i]) + '\,' + str(L[i]) + '}$')
    params_names = np.array(params_names)

    params_used = [int(i) for i in matrix[:1][0]]
    _matrix = matrix[1:]

    x_ticks, y_ticks = _matrix.shape

    extent = [0, x_ticks, 0, y_ticks]

    if diag:
        k = -1
        # idx represents the ignored elements
        labels_x = params_names[params_used]
        labels_y = labels_x[::-1]
    else:
        k = 0
        # idx represents the ignored elements
        labels_x = params_names[params_used][:-1]
        labels_y = labels_x[::-1][:-1]

    # selecting half covariance
    mask = np.tri(_matrix.shape[0], k=k)
    matrix_mask = np.ma.array(_matrix, mask=mask).T
    # mask out the lower triangle

    fig, ax = plt.subplots()

    # get rid of the frame
    for spine in plt.gca().spines.values():
        spine.set_visible(False)

    im = ax.imshow(
        X=matrix_mask,
        extent=extent,
        vmax=_matrix.max(),
        vmin=_matrix.min(),
        cmap=plt.cm.Reds,
        interpolation='nearest',
        origin='upper'
        )

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="3%", pad=0.03)
    cb = fig.colorbar(im, cax=cax)
    cb.formatter.set_powerlimits((0, 0))
    cb.ax.yaxis.set_offset_position('left')
    cb.update_ticks()
    cb.ax.set_ylabel(cbtitle)

    ax.set_title(title)
    ax.set_xticks(np.arange(x_ticks) + 0.5)
    ax.set_xticklabels(labels_x, rotation='vertical')
    ax.set_yticks(np.arange(y_ticks) + 0.5)
    ax.set_yticklabels(labels_y)
    ax.grid('off')

    fig.tight_layout()

    return fig


def plot_fit_path(
    path_pyoof, order, illum_func, telgeo, resolution, box_factor, angle,
    plim_rad, save
        ):
    """
    Plot all important figures after a least squares minimization.

    Parameters
    ----------
    path_pyoof : `str`
        Path to the pyoof output, ``'pyoof_out/directory'``.
    order : `int`
        Order used for the Zernike circle polynomial, :math:`n`.
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with the key **I_coeff**. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.
    telgeo : `list`
        List that contains the blockage distribution, optical path difference
        (OPD) function, and the primary radius (`float`) in meters. The list
        must have the following order, ``telego = [block_dist, opd_func, pr]``.
    resolution : `int`
        Fast Fourier Transform resolution for a rectangular grid. The input
        value has to be greater or equal to the telescope resolution and with
        power of 2 for faster FFT processing. It is recommended a value higher
        than ``resolution = 2 ** 8``.
    box_factor : `int`
        Related to the FFT resolution (**resolution** key), defines the image
        pixel size level. It depends on the primary radius, ``pr``, of the
        telescope, e.g. a ``box_factor = 5`` returns ``x = np.linspace(-5 *
        pr, 5 * pr, resolution)``, an array to be used in the FFT2
        (`~numpy.fft.fft2`).
    angle : `str`
        Angle unit, it can be ``'degrees'`` or ``'radians'``.
    plim_rad : `~numpy.ndarray`
        Contains the maximum values for the :math:`u` and :math:`v`
        wave-vectors, it can be in degrees or radians depending which one is
        chosen in **angle** key. The `~numpy.ndarray` must be in the following
        order, ``plim_rad = np.array([umin, umax, vmin, vmax])``.
    save : `bool`
        If ``True``, it stores all plots in the ``'pyoof_out/directory'``
        directory.

    Returns
    -------
    fig_beam : `~matplotlib.figure.Figure`
        The three beam maps plotted from the input parameters. Each map with a
        different offset :math:`d_z` value. From left to right, :math:`d_z^-`,
        :math:`0` and :math:`d_z^+`.
    fig_phase : `~matplotlib.figure.Figure`
        Aperture phase distribution for the Zernike circle polynomials for the
        telescope primary dish.
    fig_res : `~matplotlib.figure.Figure`
        Figure from the three observed beam maps residual. Each map with a
        different offset :math:`d_z` value. From left to right, :math:`d_z^-`,
        :math:`0` and :math:`d_z^+`.
    fig_data : `~matplotlib.figure.Figure`
        Figure from the three observed beam maps. Each map with a different
        offset :math:`d_z` value. From left to right, :math:`d_z^-`, :math:`0`
        and :math:`d_z^+`.
    fig_cov : `~matplotlib.figure.Figure`
        Triangle figure representing Variance-Covariance matrix.
    fig_corr : `~matplotlib.figure.Figure`
        Triangle figure representing Correlation matrix.
    """

    try:
        path_pyoof
    except NameError:
        print('pyoof directory does not exist: ' + path_pyoof)
    else:
        pass

    if not os.path.exists(path_pyoof + '/plots'):
        os.makedirs(path_pyoof + '/plots')

    path_plot = path_pyoof + '/plots'

    # Info
    n = order
    fitpar = ascii.read(path_pyoof + '/fitpar_n' + str(n) + '.csv')
    K_coeff = np.array(fitpar['parfit'])[4:]

    with open(path_pyoof + '/pyoof_info.yml', 'r') as inputfile:
        pyoof_info = yaml.load(inputfile)

    obs_object = pyoof_info['obs_object']
    meanel = round(pyoof_info['meanel'], 2)

    # Residual
    res = np.genfromtxt(path_pyoof + '/res_n' + str(n) + '.csv')

    # Data
    u_data = np.genfromtxt(path_pyoof + '/u_data.csv')
    v_data = np.genfromtxt(path_pyoof + '/v_data.csv')
    beam_data = np.genfromtxt(path_pyoof + '/beam_data.csv')

    # Covariance and Correlation matrix
    cov = np.genfromtxt(path_pyoof + '/cov_n' + str(n) + '.csv')
    corr = np.genfromtxt(path_pyoof + '/corr_n' + str(n) + '.csv')

    if n == 1:
        fig_data = plot_data(
            u_data=u_data,
            v_data=v_data,
            beam_data=beam_data,
            d_z=pyoof_info['d_z'],
            title='{} observed power pattern $\\alpha={}$ degrees'.format(
                obs_object, meanel),
            angle=angle,
            res_mode=False
            )

    fig_beam = plot_beam(
        params=np.array(fitpar['parfit']),
        title='{} fit power pattern  $n={}$ $\\alpha={}$ degrees'.format(
            obs_object, n, meanel),
        d_z=pyoof_info['d_z'],
        wavel=pyoof_info['wavel'],
        illum_func=illum_func,
        telgeo=telgeo,
        plim_rad=plim_rad,
        angle=angle,
        resolution=resolution,
        box_factor=box_factor
        )

    fig_phase = plot_phase(
        K_coeff=K_coeff,
        title=(
            '{} phase error $d_z=\\pm {}$ m ' +
            '$n={}$ $\\alpha={}$ deg'
            ).format(obs_object, round(pyoof_info['d_z'][2], 3), n, meanel),
        notilt=True,
        pr=telgeo[2]
        )

    fig_res = plot_data(
        u_data=u_data,
        v_data=v_data,
        beam_data=res,
        d_z=pyoof_info['d_z'],
        title='{} residual $n={}$'.format(obs_object, n),
        angle=angle,
        res_mode=True
        )

    fig_cov = plot_variance(
        matrix=cov,
        order=n,
        title='{} variance-covariance matrix $n={}$'.format(obs_object, n),
        cbtitle='$\sigma_{ij}^2$',
        diag=True
        )

    fig_corr = plot_variance(
        matrix=corr,
        order=n,
        title='{} correlation matrix $n={}$'.format(obs_object, n),
        cbtitle='$\\rho_{ij}$',
        diag=True
        )

    if save:
        fig_beam.savefig(path_plot + '/fitbeam_n{}.pdf'.format(n))
        fig_phase.savefig(path_plot + '/fitphase_n{}.pdf'.format(n))
        fig_res.savefig(path_plot + '/residual_n{}.pdf'.format(n))
        fig_cov.savefig(path_plot + '/cov_n{}.pdf'.format(n))
        fig_corr.savefig(path_plot + '/corr_n{}.pdf'.format(n))

        if n == 1:
            fig_data.savefig(path_plot + '/obsbeam.pdf')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import numpy as np
import os
from scipy.constants import c as light_speed
from astropy.io import fits
from .aperture import radiation_pattern
from .math_functions import wavevector2radians

__all__ = ['beam_generator']


def beam_generator(
    params, wavel, d_z, illum_func, telgeo, noise, resolution, box_factor
        ):
    """
    Routine to generate data and test the pyoof package algorithm. It has the
    default setting for the pyoof fits file input.

    Parameters
    ----------
    params : `~numpy.ndarray`
        Two stacked arrays, the illumination and Zernike circle polynomials
        coefficients. ``params = np.hstack([I_coeff, K_coeff])``.
    wavel : `float`
        Wavelength, :math:`\\lambda`, in meters.
    d_z : `list`
        Radial offset :math:`d_z`, added to the sub-reflector in meters. This
        characteristic measurement adds the classical interference pattern to
        the beam maps, normalized squared (field) radiation pattern, which is
        an out-of-focus property. The radial offset list must be as follows,
        ``d_z = [d_z-, 0., d_z+]`` all of them in meters.
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with the key **I_coeff**. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.
    telgeo : `list`
        List that contains the blockage distribution, optical path difference
        (OPD) function, and the primary radius (`float`) in meters. The list
        must have the following order, ``telego = [block_dist, opd_func, pr]``.
    noise : `float`
        Noise amplitude added to the generated data. The noise comes from a
        random Gaussian, see `~numpy.random.normal`.
    resolution : `int`
        Fast Fourier Transform resolution for a rectangular grid. The input
        value has to be greater or equal to the telescope resolution and with
        power of 2 for faster FFT processing. It is recommended a value higher
        than ``resolution = 2 ** 8``.
    box_factor : `int`
        Related to the FFT resolution (**resolution** key), defines the image
        pixel size level. It depends on the primary radius, ``pr``, of the
        telescope, e.g. a ``box_factor = 5`` returns ``x = np.linspace(-5 *
        pr, 5 * pr, resolution)``, an array to be used in the FFT2
        (`~numpy.fft.fft2`).

    Returns
    -------
    pyoof_fits : `~astropy.io.fits.hdu.hdulist.HDUList`
        The output fits file is stored in the directory ``'data_generated/'``.
        Every time the function is executed a new file will be stored. The
        file is ready to use for the `~pyoof` package.
    """

    freq = light_speed / wavel  # Hz frequency
    bw = 1.22 * wavel / (2 * telgeo[2])  # Beamwidth
    size_in_bw = 8 * bw
    plim_u = [-size_in_bw, size_in_bw]  # radians
    plim_v = [-size_in_bw, size_in_bw]

    I_coeff, K_coeff = params[:4], params[4:]

    # Generating power pattern and spatial frequencies
    u, v, P = [], [], []
    for i in range(3):
        _u, _v, _radiation = radiation_pattern(
            K_coeff=K_coeff,
            I_coeff=I_coeff,
            d_z=d_z[i],
            wavel=wavel,
            illum_func=illum_func,
            telgeo=telgeo,
            resolution=resolution,
            box_factor=box_factor
            )

        _u = wavevector2radians(_u, wavel)  # radians
        _v = wavevector2radians(_v, wavel)

        uu, vv = np.meshgrid(_u, _v)
        power_pattern = np.abs(_radiation) ** 2

        # trim the power pattern
        power_pattern[plim_u[0] > uu] = np.nan
        power_pattern[plim_u[1] < uu] = np.nan
        power_pattern[plim_v[0] > vv] = np.nan
        power_pattern[plim_v[1] < vv] = np.nan

        power_trim_1d = power_pattern[~np.isnan(power_pattern)]
        size_trim = int(np.sqrt(power_trim_1d.size))  # new size of the box

        # Box to be trimmed in uu and vv meshed arrays
        box = [(plim_u[0] < uu) & (plim_u[1] > uu) & (plim_v[0] < vv) & (
            plim_v[1] > vv)]

        # reshaping trimmed arrys
        power_trim = power_trim_1d.reshape(size_trim, size_trim)
        u_trim = uu[box].reshape(size_trim, size_trim)
        v_trim = vv[box].reshape(size_trim, size_trim)

        u.append(u_trim)
        v.append(v_trim)
        P.append(power_trim)

    # adding noise!
    if noise == 0:
        power_noise = np.array(P)
    else:
        power_noise = (
            np.array(P) + np.random.normal(0., noise, np.array(P).shape)
            )
    power_norm = [power_noise[i] / power_noise[i].max() for i in range(3)]

    u_to_save = [u[i].flatten() for i in range(3)]
    v_to_save = [v[i].flatten() for i in range(3)]
    p_to_save = [power_norm[i].flatten() for i in range(3)]

    # Writing default fits file for OOF observations
    table_hdu0 = fits.BinTableHDU.from_columns([
        fits.Column(name='U', format='E', array=u_to_save[0]),
        fits.Column(name='V', format='E', array=v_to_save[0]),
        fits.Column(name='BEAM', format='E', array=p_to_save[0])
        ])

    table_hdu1 = fits.BinTableHDU.from_columns([
        fits.Column(name='U', format='E', array=u_to_save[1]),
        fits.Column(name='V', format='E', array=v_to_save[1]),
        fits.Column(name='BEAM', format='E', array=p_to_save[1])
        ])

    table_hdu2 = fits.BinTableHDU.from_columns([
        fits.Column(name='U', format='E', array=u_to_save[2]),
        fits.Column(name='V', format='E', array=v_to_save[2]),
        fits.Column(name='BEAM', format='E', array=p_to_save[2])
        ])

    table_hdu0.update_ext_name('MINUS OOF')
    table_hdu1.update_ext_name('ZERO OOF')
    table_hdu2.update_ext_name('PLUS OOF')

    # storing data
    current_path = os.getcwd()
    if not os.path.exists(current_path + '/data_generated'):
        os.makedirs(current_path + '/data_generated')

    for j in ["%03d" % i for i in range(101)]:
        name_file = current_path + '/data_generated/test{}.fits'.format(j)
        if not os.path.exists(name_file):

            prihdr = fits.Header()
            prihdr['FREQ'] = freq
            prihdr['WAVEL'] = wavel
            prihdr['MEANEL'] = 0
            prihdr['OBJECT'] = 'test{}'.format(j)
            prihdr['DATE_OBS'] = 'test{}'.format(j)
            prihdr['COMMENT'] = 'Generated data pyoof package'
            prihdr['AUTHOR'] = 'Tomas Cassanelli'
            prihdu = fits.PrimaryHDU(header=prihdr)
            pyoof_fits = fits.HDUList(
                [prihdu, table_hdu0, table_hdu1, table_hdu2]
                )

            for i in range(3):
                pyoof_fits[i + 1].header['DZ'] = d_z[i]

            pyoof_fits.writeto(name_file, clobber=True)

            break

    return pyoof_fits



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import numpy as np
from math import factorial as f


__all__ = [
    'U', 'R'
    ]


def R(n, m, rho):
    """
    Radial Zernike polynomials generator (:math:`R^m_n(\\varrho)` from Born &
    Wolf definition). The :math:`m`, :math:`n` are integers, :math:`n\geqslant
    0` and :math:`n - m` even. Only used to compute the general expression for
    the Zernike circle polynomials, `~pyoof.zernike.U`.

    Parameters
    ----------
    n : `int`
        It is :math:`n \geqslant 0`. Order of the radial component.
    m : `int`
        Positive number, relative to the angle component.
    rho : `~numpy.ndarray`
        Values for the radial component, :math:`\\varrho = \sqrt{x^2 + y^2}`.

    Returns
    -------
    radial_poly : `~numpy.ndarray`
        Radial Zernike polynomial already evaluated, :math:`R^m_n(\\varrho)`.

    Notes
    -----
    The original generating formula for the radial polynomials is given by,

    .. math::

        R^{\\pm m}_n (\\varrho) = \\frac{1}{\\left(\\frac{n-m}{2}\\right)!\cdot
        \\varrho^m}\\left\{\\frac{\\mathrm{d}}{\\mathrm{d}\\left(\\varrho^2
        \\right)} \\right\}^{\\frac{n-m}{2}} \\left\{  \\left( \\varrho^2
        \\right)^{\\frac{n+m}{2}} \cdot  \\left( \\varrho^2 -1
        \\right)^{\\frac{n-m}{2}} \\right\},

    Which can also be expressed as a polynomial sum.
    """

    a = (n + m) // 2
    b = (n - m) // 2

    radial_poly = sum(
        (-1) ** s * f(n - s) * rho ** (n - 2 * s) /
        (f(s) * f(a - s) * f(b - s))
        for s in range(0, b + 1)
        )

    return radial_poly


def U(n, l, rho, theta):
    """
    Zernike circle polynomials generator (:math:`U^\ell_n(\\varrho, \\varphi)`
    from Born & Wolf definition). The :math:`\ell`, :math:`n` are integers,
    :math:`n \geqslant 0` and :math:`n - |\ell|` even. Expansion of a complete
    set of orthonormal polynomials in a unitary circle, specially useful when
    computing for the wavefront (aberration) distribution, :math:`W(x, y)`.
    The total number of polynomials is given by :math:`(n + 1)(n + 2) / 2.`

    Parameters
    ----------
    n : `int`
        It is :math:`n \geqslant 0`. Relative to radial component.
    l : `int`
        l Can be positive or negative, relative to angle component.
    rho : `~numpy.ndarray`
        Values for the radial component. :math:`\\varrho = \\sqrt{x^2 + y^2}`.
    theta : `~numpy.ndarray`
        Values for the angular component. For a rectangular grid x and y are
        evaluated as :math:`\\vartheta = \\mathrm{arctan}(y / x)`.

    Returns
    -------
    zernike_circle_poly : `~numpy.ndarray`
        Zernike circle polynomial already evaluated, :math:`U^\ell_n(\\varrho,
        \\varphi)`.

    Notes
    -----
    The generating formula for the Zernike circle polymials make use of the
    radial polynomials, then,

    .. math::
        U^\ell_n(\\varrho, \\vartheta) = R^m_n(\\varrho) \\cdot \\cos
        m\\vartheta \qquad \\ell \geq 0,

    .. math::

        U^\ell_n(\\varrho, \\vartheta) = R^m_n(\\varrho) \\cdot \sin
        m\\vartheta \\qquad \\ell < 0.
    """

    if not isinstance(l, int):
        raise TypeError(
            'Polynomial angular component (l) has to be an integer'
            )
    if not (n >= 0 and isinstance(n, int)):
        raise TypeError('Polynomial order (n) has to be a positive integer')

    m = abs(l)
    radial = R(n, m, rho)

    if l < 0:
        zernike_circle_poly = radial * np.sin(m * theta)
    else:
        zernike_circle_poly = radial * np.cos(m * theta)

    return zernike_circle_poly



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import pytest
import numpy as np
from astropy.utils.misc import NumpyRNGContext
from astropy.utils.data import get_pkg_data_filename
from numpy.testing import assert_equal, assert_allclose
import pyoof


def test_cart2pol():

    with NumpyRNGContext(0):
        x = np.random.uniform(-1, 1, 5)
        y = np.random.uniform(-1, 1, 5)

    rho, theta = pyoof.cart2pol(x, y)

    rho_true = np.array([
        0.30768718, 0.44811525, 0.81005283, 0.93166014, 0.27867163
        ])
    theta_true = np.array([
        1.24792264, -0.28229117, 1.31427197, 1.47429564, -2.15067483
        ])

    assert_allclose(rho, rho_true)
    assert_allclose(theta, theta_true)


def test_wavevector2degrees():

    wavel = np.linspace(0.005, 0.014, 10)

    with NumpyRNGContext(0):
        u = np.sort(np.random.uniform(-2.5, 2.5, 10))

    wavevector_degrees = pyoof.wavevector2degrees(u, wavel)

    wavevector_degrees_true = np.array([
        -0.16695773, -0.13122773, -0.12515963, 0.10286468, 0.12585635,
        0.29439539, 0.45975143, 0.73976655, 1.45905107, 1.85961435
        ])

    assert_allclose(wavevector_degrees, wavevector_degrees_true)


def test_wavevector2radians():

    wavel = np.linspace(0.005, 0.014, 10)

    with NumpyRNGContext(0):
        u = np.sort(np.random.uniform(-2.5, 2.5, 10))

    wavevector_radians = pyoof.wavevector2radians(u, wavel)

    wavevector_radians_true = np.array([
        -0.002913962029, -0.00229035602, -0.002184447606, 0.00179532732,
        0.002196607677, 0.005138168804, 0.008024176219, 0.012911361982,
        0.025465245051, 0.032456393235
        ])

    assert_allclose(wavevector_radians, wavevector_radians_true)


def test_co_matrices():

    # importing tests files
    res = np.load(
        get_pkg_data_filename('data/res_co_matrices.npy')
        )
    jac = np.load(
        get_pkg_data_filename('data/jac_co_matrices.npy')
        )
    cov_true = np.load(
        get_pkg_data_filename('data/cov_co_matrices.npy')
        )
    corr_true = np.load(
        get_pkg_data_filename('data/corr_co_matrices.npy')
        )

    cov, corr = pyoof.co_matrices(res, jac, 1)

    assert_allclose(cov, cov_true)
    assert_allclose(corr, corr_true)


def test_line_equation():

    x = np.linspace(-1, 1, 10)
    y1 = pyoof.line_equation((0, 0), (1, 1,), x)
    y1_true = x

    with NumpyRNGContext(0):
        p1 = (np.random.uniform(0, 1), np.random.uniform(0, 1))
        p2 = (np.random.uniform(0, 1), np.random.uniform(0, 1))

    y2 = pyoof.line_equation(p1, p2, x)
    y2_true = np.array([
        5.604404226951, 4.902904538006, 4.20140484906, 3.499905160114,
        2.798405471169, 2.096905782223, 1.395406093278, 0.693906404332,
        -0.007593284613, -0.709092973559
        ])

    assert_allclose(y1, y1_true)
    assert_allclose(y2, y2_true)


def test_rms():

    x1 = np.array([1] * 10 + [-1] * 10)
    rms1 = pyoof.rms(x1)
    rms1_true = 1.0

    with NumpyRNGContext(0):
        x2 = np.random.uniform(-20, 20, 5)
    rms2 = pyoof.rms(x2)
    rms2_true = 4.6335342124813295

    assert_equal(rms1, rms1_true)
    assert_equal(rms2, rms2_true)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import numpy as np
from ..math_functions import cart2pol, rms
from ..zernike import U

__all__ = [
    'illum_pedestal', 'illum_gauss', 'wavefront', 'phase', 'aperture',
    'radiation_pattern', 'e_rs'
    ]


def e_rs(phase):
    """
    Computes the random-surface-error efficiency,
    :math:`\\varepsilon_\\mathrm{rs}`, using the Ruze's equation.

    Parameters
    ----------
    phase : `~numpy.ndarray`
        The phase error, :math:`\\varphi(x, y)`, is a two dimensional array (
        one of the solutions from the pyoof package). Its amplitude values are
        in radians.

    Notes
    -----
    Ruze's equation is derived empirically from a fat reflector with Gaussian
    distributed errors, and it expressed as,

    .. math::
        \\varepsilon_\\mathrm{rs} =
        \\mathrm{e}^{-(4\pi\delta_\\mathrm{rms}/\lambda)^2}.

    Where :math:`\delta_\\mathrm{rms}` corresponds to the root-mean-squared
    deviation. The Python function uses the key **phase** because the term
    :math:`4\pi\delta_\\mathrm{rms}/\lambda` corresponds to the phase error.
    """

    rms_rad = rms(phase)  # rms value in radians

    return np.exp(-rms_rad ** 2)


def illum_pedestal(x, y, I_coeff, pr, q=2):
    """
    Illumination function, :math:`E_\\mathrm{a}(x, y)`, parabolic taper on a
    pedestal, sometimes called apodization, taper or window. Represents the
    distribution of light in the primary reflector. The illumination reduces
    the side lobes in the FT and it is a property of the
    receiver.

    Parameters
    ----------
    x : `~numpy.ndarray`
        Grid value for the :math:`x` variable in meters.
    y : `~numpy.ndarray`
        Grid value for the :math:`y` variable in meters.
    I_coeff : `~numpy.ndarray`
        List which contains 4 parameters, the illumination amplitude,
        :math:`A_{E_\\mathrm{a}}`, the illumination taper,
        :math:`c_\\mathrm{dB}` and the two coordinate offset, :math:`(x_0,
        y_0)`. The illumination coefficients must be listed as follows,
        ``I_coeff = [i_amp, c_dB, x0, y0]``.
    pr : `float`
        Primary reflector radius in meters.
    q : `int`
        Order of the parabolic taper on a pedestal, it is commonly set at
        :math:`q = 2`.

    Returns
    -------
    Ea : `~numpy.ndarray`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`.

    Notes
    -----
    The parabolic taper on a pedestal has the following mathematical formula,

    .. math::

        E_{\\mathrm{a}} (\\rho') = C + (1 - C) \\cdot
        \\left[ 1 - \\left( \\frac{\\rho'}{R} \\right)^2 \\right]^q,

    .. math::

        C=10^{\\frac{c_\\mathrm{dB}}{20}}.

    """

    i_amp, c_dB, x0, y0 = I_coeff

    c = 10 ** (c_dB / 20.)  # c_dB has to be negative, bounds given [-8, -25]
    r = np.sqrt((x - x0) ** 2 + (y - y0) ** 2)

    # Parabolic taper on a pedestal
    Ea = i_amp * (c + (1. - c) * (1. - (r / pr) ** 2) ** q)

    return Ea


def illum_gauss(x, y, I_coeff, pr):
    """
    Illumination function, :math:`E_\\mathrm{a}(x, y)`, Gaussian, sometimes
    called apodization, taper or window. Represents the distribution of light
    in the primary reflector. The illumination reduces the side lobes in the
    FT and it is a property of the receiver.

    Parameters
    ----------
    x : `~numpy.ndarray`
        Grid value for the :math:`x` variable in meters.
    y : `~numpy.ndarray`
        Grid value for the :math:`y` variable in meters.
    I_coeff : `~numpy.ndarray`
        List which contains 4 parameters, the illumination amplitude,
        :math:`A_{E_\\mathrm{a}}`, the illumination taper,
        :math:`\\sigma_\\mathrm{dB}` and the two coordinate offset, :math:`(
        x_0, y_0)`. The illumination coefficients must be listed as follows,
        ``I_coeff = [i_amp, sigma_dB, x0, y0]``.
    pr : `float`
        Primary reflector radius in meters.

    Returns
    -------
    Ea : `~numpy.ndarray`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`.

    Notes
    -----
    The Gaussian illumination function has the same formula of a normalized
    Gaussian distribution.
    """
    i_amp, sigma_dB, x0, y0 = I_coeff
    sigma = 10 ** (sigma_dB / 20)  # -15 to -20 dB
    norm = np.sqrt(2 * np.pi * sigma ** 2)  # normalization Gaussian
    Ea = (
        i_amp * norm *
        np.exp(-((x - x0) ** 2 + (y - y0) ** 2) / (2 * (sigma * pr) ** 2))
        )

    return Ea


def wavefront(rho, theta, K_coeff):
    """
    Computes the wavefront (aberration) distribution, :math:`W(x, y)`. It
    tells how is the error distributed along the primary dish, it is related
    to the phase error. The wavefront (aberration) distribution is described
    as a parametrization of the Zernike circle polynomials multiplied by a set
    of coefficients, :math:`K_{n\\ell}`.

    Parameters
    ----------
    rho : `~numpy.ndarray`
        Values for the radial component. :math:`\sqrt{x^2 + y^2} /
        \\varrho_\\mathrm{max}` normalized by its maximum radius.
    theta : `~numpy.ndarray`
        Values for the angular component. :math:`\\vartheta = \mathrm{arctan}(
        y / x)`.
    K_coeff : `~numpy.ndarray`
        Constants coefficients, :math:`K_{n\\ell}`, for each of them there is
        only one Zernike circle polynomial, :math:`U^\ell_n(\\varrho,
        \\varphi)`. The coefficients are between :math:`[-2, 2]`.

    Returns
    -------
    W : `~numpy.ndarray`
        Wavefront (aberration) distribution, :math:`W(x, y)`. Zernike circle
        polynomials already evaluated and multiplied by their coefficients.

    Notes
    -----
    The wavefront (aberration) distribution it strictly related to the Zernike
    circle polynomials through the expression,

    .. math::
        W(\\varrho, \\vartheta) = \\sum_{n,
        \\ell}K_{n\\ell}U^\\ell_n(\\varrho, \\vartheta).
    """

    # Total number of Zernike circle polynomials
    n = int((np.sqrt(1 + 8 * K_coeff.size) - 3) / 2)

    # list of tuples with (n, l) allowed values
    nl = [(i, j) for i in range(0, n + 1) for j in range(-i, i + 1, 2)]

    # Wavefront (aberration) distribution
    W = sum(
        K_coeff[i] * U(*nl[i], rho, theta)
        for i in range(K_coeff.size)
        )

    return W


def phase(K_coeff, notilt, pr, resolution=1e3):
    """
    Aperture phase distribution (or phase error), :math:`\\varphi(x, y)`, for
    an specific telescope primary reflector. In general the tilt (average
    slope in :math:`x`- and :math:`y`-directions, related to telescope
    pointings) is subtracted from its calculation. Function used to show the
    final results from the fit procedure.

    Parameters
    ----------
    K_coeff : `~numpy.ndarray`
        Constants coefficients, :math:`K_{n\\ell}`, for each of them there is
        only one Zernike circle polynomial, :math:`U^\ell_n(\\varrho,
        \\varphi)`. The coefficients are between :math:`[-2, 2]`.
    notilt : `bool`
        Boolean to include or exclude the tilt coefficients in the aperture
        phase distribution. The Zernike circle polynomials are related to tilt
        through :math:`U^{-1}_1(\\varrho, \\varphi)` and
        :math:`U^1_1(\\varrho, \\varphi)`.
    pr : `float`
        Primary reflector radius in meters.
    resolution : `int`
        Resolution for the phase error map, usually used ``resolution = 1e3``
        in the pyoof package.

    Returns
    -------
    x : `~numpy.ndarray`
        :math:`x`-axis dimensions for the primary reflector in meters.
    y : `~numpy.ndarray`
        :math:`y`-axis dimensions for the primary reflector in meters.
    phi : `~numpy.ndarray`
        Aperture phase distribution, :math:`\\varphi(x, y)`, for an specific
        primary dish radius, measured in radians.

    Notes
    -----
    The aperture phase distribution or phase error, :math:`\\varphi(x, y)` is
    related to the wavefront (aberration) distribution, from classical optics,
    through the expression,

    .. math::
        \\varphi(x, y) = 2\pi \\cdot W(x, y) = 2\pi \\cdot \\sum_{n, \\ell}
        K_{n\\ell}U^\\ell_n(\\varrho, \\vartheta).
    """

    _K_coeff = K_coeff.copy()

    # Erasing tilt dependence
    if notilt:
        _K_coeff[1] = 0  # For value K(-1, 1) = 0
        _K_coeff[2] = 0  # For value K(1, 1) = 0

    x = np.linspace(-pr, pr, resolution)
    y = np.linspace(-pr, pr, resolution)
    x_grid, y_grid = np.meshgrid(x, y)

    r, t = cart2pol(x_grid, y_grid)
    r_norm = r / pr  # For orthogonality U(n, l) polynomials

    # Wavefront (aberration) distribution
    W = wavefront(rho=r_norm, theta=t, K_coeff=_K_coeff)
    W[(x_grid ** 2 + y_grid ** 2 > pr ** 2)] = 0

    phi = W * 2 * np.pi  # Aperture phase distribution in radians

    return x, y, phi


def aperture(x, y, K_coeff, I_coeff, d_z, wavel, illum_func, telgeo):
    """
    Aperture distribution, :math:`\\underline{E_\\mathrm{a}}(x, y)`.
    Collection of individual distribution/functions: i.e. illumination
    function, :math:`E_\\mathrm{a}(x, y)`, blockage distribution, :math:`B(x,
    y)`, aperture phase distribution, :math:`\\varphi(x, y)` and OPD
    function, :math:`\\delta(x, y;d_z)`. In general, it is a complex
    quantity, its phase an amplitude are better understood separately. The FT
    (2-dim) of the aperture represents the (field) radiation pattern, :math:`F(
    u, v)`.

    Parameters
    ----------
    x : `~numpy.ndarray`
        Grid value for the :math:`x` variable in meters.
    y : `~numpy.ndarray`
        Grid value for the :math:`y` variable in meters.
    K_coeff : `~numpy.ndarray`
        Constants coefficients, :math:`K_{n\\ell}`, for each of them there is
        only one Zernike circle polynomial, :math:`U^\ell_n(\\varrho,
        \\varphi)`. The coefficients are between :math:`[-2, 2]`.
    I_coeff : `~numpy.ndarray`
        List which contains 4 parameters, the illumination amplitude,
        :math:`A_{E_\\mathrm{a}}`, the illumination taper,
        :math:`c_\\mathrm{dB}` and the two coordinate offset, :math:`(x_0,
        y_0)`. The illumination coefficients must be listed as follows,
        ``I_coeff = [i_amp, c_dB, x0, y0]``.
    d_z : `float`
        Radial offset, :math:`d_z`, added to the sub-reflector in meters. This
        characteristic measurement adds the classical interference pattern to
        the beam maps, normalized squared (field) radiation pattern, which is
        an out-of-focus property. It is usually of the order of centimeters.
    wavel : `float`
        Wavelength, :math:`\\lambda`, of the observation in meters.
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with the key **I_coeff**. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.
    telgeo : `list`
        List that contains the blockage distribution, optical path difference
        (OPD) function, and the primary radius (`float`) in meters. The list
        must have the following order, ``telego = [block_dist, opd_func, pr]``.

    Returns
    -------
    E : `~numpy.ndarray`
        Grid value that contains general expression for aperture distribution,
        :math:`\\underline{E_\\mathrm{a}}(x, y)`.

    Notes
    -----
    The aperture distribution is a collection of distributions/functions, its
    structure follows,

    .. math::
        \\underline{E_\\mathrm{a}}(x, y) = B(x, y)\\cdot E_\\mathrm{a}(x, y)
        \\cdot \\mathrm{e}^{\\mathrm{i} \{\\varphi(x, y) +
        \\frac{2\pi}{\lambda}\\delta(x,y;d_z)\}}.

    Where it does represent a complex number, with phase: aperture phase
    distribution, plus OPD function and amplitude the blockage distribution
    and illumination function.

    """

    r, t = cart2pol(x, y)

    [block_dist, opd_func, pr] = telgeo
    B = block_dist(x=x, y=y)

    # Normalization to be used in the Zernike circle polynomials
    r_norm = r / pr

    # Wavefront (aberration) distribution
    W = wavefront(rho=r_norm, theta=t, K_coeff=K_coeff)
    delta = opd_func(x=x, y=y, d_z=d_z)  # Optical path difference function
    Ea = illum_func(x=x, y=y, I_coeff=I_coeff, pr=pr)  # Illumination function

    # Transformation: wavefront (aberration) distribution -> phase error
    phi = (W + delta / wavel) * 2 * np.pi  # phase error plus the OPD function

    E = B * Ea * np.exp(phi * 1j)  # Aperture distribution

    return E


def radiation_pattern(
    K_coeff, I_coeff, d_z, wavel, illum_func, telgeo, resolution, box_factor
        ):
    """
    Spectrum or (field) radiation pattern, :math:`F(u, v)`, it is the FFT2
    computation of the aperture distribution, :math:`\\underline{E_\\mathrm{a}}
    (x, y)`, in a rectangular grid. Passing the majority of
    arguments to the aperture distribution except the FFT2 resolution.

    Parameters
    ----------
    K_coeff : `~numpy.ndarray`
        Constants coefficients, :math:`K_{n\\ell}`, for each of them there is
        only one Zernike circle polynomial, :math:`U^\ell_n(\\varrho,
        \\varphi)`. The coefficients are between :math:`[-2, 2]`.
    I_coeff : `~numpy.ndarray`
        List which contains 4 parameters, the illumination amplitude,
        :math:`A_{E_\\mathrm{a}}`, the illumination taper,
        :math:`c_\\mathrm{dB}` and the two coordinate offset, :math:`(x_0,
        y_0)`. The illumination coefficients must be listed as follows,
        ``I_coeff = [i_amp, c_dB, x0, y0]``.
    d_z : `float`
        Radial offset, :math:`d_z`, added to the sub-reflector in meters. This
        characteristic measurement adds the classical interference pattern to
        the beam maps, normalized squared (field) radiation pattern, which is
        an out-of-focus property. It is usually of the order of centimeters.
    wavel : `float`
        Wavelength, :math:`\\lambda`, of the observation in meters.
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with the key **I_coeff**. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.
    telgeo : `list`
        List that contains the blockage distribution, optical path difference
        (OPD) function, and the primary radius (`float`) in meters. The list
        must have the following order, ``telego = [block_dist, opd_func, pr]``.
    resolution : `int`
        Fast Fourier Transform resolution for a rectangular grid. The input
        value has to be greater or equal to the telescope resolution and with
        power of 2 for faster FFT processing. It is recommended a value higher
        than ``resolution = 2 ** 8``.
    box_factor : `int`
        Related to the FFT resolution (**resolution** key), defines the image
        pixel size level. It depends on the primary radius, ``pr``, of the
        telescope, e.g. a ``box_factor = 5`` returns ``x = np.linspace(-5 *
        pr, 5 * pr, resolution)``, an array to be used in the FFT2
        (`~numpy.fft.fft2`).

    Returns
    -------
    u_shift : `~numpy.ndarray`
        :math:`u` wave-vector in 1 / m units. It belongs to the :math:`x`
        coordinate in meters from the aperture distribution,
        :math:`\\underline{E_\\mathrm{a}}(x, y)`.
    v_shift : `~numpy.ndarray`
        :math:`v` wave-vector in 1 / m units. It belongs to the y coordinate
        in meters from the aperture distribution,
        :math:`\\underline{E_\\mathrm{a}}(x, y)`.
    F_shift : `~numpy.ndarray`
        Output from the FFT2 (`~numpy.fft.fft2`), :math:`F(u, v)`,
        unnormalized solution in a grid, defined by **resolution** and
        **box_factor** keys.

    Notes
    -----
    The (field) radiation pattern is the direct Fourier Transform in two
    dimensions of the aperture distribution, hence,

    .. math::

        F(u, v) = \\mathcal{F} \\left[ \\underline{E_\\mathrm{a}}(x, y)
        \\right].
    """

    # Arrays to generate (field) radiation pattern
    pr = telgeo[2]
    box_size = pr * box_factor

    x = np.linspace(-box_size, box_size, resolution)
    y = x
    x_grid, y_grid = np.meshgrid(x, y)

    dx = x[1] - x[0]
    dy = y[1] - y[0]

    # Aperture distribution model
    E = aperture(
        x=x_grid,
        y=y_grid,
        K_coeff=K_coeff,
        I_coeff=I_coeff,
        d_z=d_z,
        wavel=wavel,
        illum_func=illum_func,
        telgeo=telgeo
        )

    F = np.fft.fft2(E)
    F_shift = np.fft.fftshift(F)  # (field) radiation pattern

    # wave-vectors in 1 / m
    u, v = np.fft.fftfreq(x.size, dx), np.fft.fftfreq(y.size, dy)
    u_shift, v_shift = np.fft.fftshift(u), np.fft.fftshift(v)

    return u_shift, v_shift, F_shift



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import pytest
import numpy as np
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.misc import NumpyRNGContext
from numpy.testing import assert_equal, assert_allclose
import pyoof


def test_e_rs():

    with NumpyRNGContext(0):
        phase = np.sort(np.random.uniform(-2.5, 2.5, (5, 5)))

    e_rs = pyoof.aperture.e_rs(phase)

    assert_allclose(e_rs, 0.110794937514)


def test_illum_pedestal():

    pr = 50
    x = np.linspace(-pr, pr)
    y = x
    xx, yy = np.meshgrid(x, y)
    I_coeff = [1, -14, 0, 0]

    _illum_pedestal = pyoof.aperture.illum_pedestal(xx, yy, I_coeff, pr)
    illum_pedestal_true = np.load(
        get_pkg_data_filename('data/illum_pedestal.npy')
        )

    assert_allclose(_illum_pedestal, illum_pedestal_true)


# def test_wavefront(rho, theta, K_coeff):

#     x = np.linspace(-50, 50, 1e3)
#     xx, yy = np.meshgrid(x, x)
#     rho, theta = pyoof.cart2pol(xx, yy)

#     with NumpyRNGContext(0):
#         K_coeff = np.sort(np.random.uniform(-2.5, 2.5, 21)


#     W = aperture.wavefront(rho, theta, K_coeff)
#     W_true = 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import numpy as np
from ..math_functions import line_equation

__all__ = [
    'opd_effelsberg', 'opd_manual', 'block_manual', 'block_effelsberg'
    ]


def opd_effelsberg(x, y, d_z):
    """
    Optical path difference (OPD) function, :math:`\\delta(x,y;d_z)`. Given by
    the geometry of the telescope and radial offset parameter, :math:`d_z`.
    This function is specific for the Effelsberg telescope.

    Parameters
    ----------
    x : `~numpy.ndarray`
        Grid value for the :math:`x` variable in meters.
    y : `~numpy.ndarray`
        Grid value for the :math:`y` variable in meters.
    d_z : `float`
        Radial offset, :math:`d_z`, added to the sub-reflector in meters. This
        characteristic measurement adds the classical interference pattern to
        the beam maps, normalized squared (field) radiation pattern, which is
        an out-of-focus property. It is usually of the order of centimeters.

    Returns
    -------
    opd : `~numpy.ndarray`
        Optical path difference function, :math:`\\delta(x,y;d_z)`.

    Notes
    -----
    For a Gregorian configuration, and with the dish diameter of Effelsberg,
    the OPD function becomes,

    .. math::
        \\delta(x,;d_z) = d_z\\left( \\frac{1-a^2}{1+a^2} +
        \\frac{1-b^2}{1+b^2} \\right),

    .. math::
        a = \\frac{\\sqrt{x^2+y^2}}{2F_\\mathrm{p}}, \\qquad b =
        \\frac{\\sqrt{x^2+y^2}}{2F_\\mathrm{eff}},

    where :math:`F_\\mathrm{p}=30` m corresponds to the parabola focal length
    and :math:`F_\\mathrm{eff}=387.4` m, the effective or total focal length.

    """

    # Cassegrain/Gregorian (at focus) telescope
    Fp = 30  # Focus primary reflector m
    F = 387.39435  # Total focus Gregorian telescope m
    r = np.sqrt(x ** 2 + y ** 2)  # polar coordinates radius
    a = r / (2 * Fp)
    b = r / (2 * F)

    opd = d_z * ((1 - a ** 2) / (1 + a ** 2) + (1 - b ** 2) / (1 + b ** 2))

    return opd


def opd_manual(Fp, F):
    """
    Optical path difference (OPD) function, :math:`\\delta(x, yd_z)`. Given by
    geometry of the telescope and defocus parameter, :math:`d_z`. For
    Cassegrain/Gregorian geometries. Primary and total (or effective) foci are
    required, :math:`F_\\mathrm{p}` and :math:`F_\\mathrm{eff}`, respectively.

    Parameters
    ----------
    Fp : `float`
        Focus primary reflector, :math:`F_\\mathrm{p}`, in meters.
    F : `float`
        Effective or total focus for the telescope mirror configuration,
        :math:`F_\\mathrm{eff}`, in meters.

    Returns
    -------
    opd_func : `function`
        It returns the function ``opd_func(x, y, d_z)``, which depends only on
        the grid and radial offset values, similar to
        `~pyoof.telgeometry.opd_effelsberg`.
    """

    def opd_func(x, y, d_z):
        # Cassegrain/Gregorian (at focus) telescope
        r = np.sqrt(x ** 2 + y ** 2)  # radial polar coordinate
        a = r / (2 * Fp)
        b = r / (2 * F)

        opd = d_z * (
            (1 - a ** 2) / (1 + a ** 2) + (1 - b ** 2) / (1 + b ** 2)
            )

        return opd

    return opd_func


def block_effelsberg(x, y):
    """
    Truncation in the aperture (amplitude) distribution, :math:`B(x, y)`,
    given by the telescope's structure; i.e. support legs, sub-reflector and
    shade effect as seen from the secondary focus of the Effelsberg telescope.

    Parameters
    ----------
    x : `~numpy.ndarray`
        Grid value for the :math:`x` variable in meters.
    y : `~numpy.ndarray`
        Grid value for the :math:`y` variable in meters.

    Returns
    -------
    block : `~numpy.ndarray`
        Aperture (amplitude) distribution truncation, :math:`B(x, y)`. Values
        that are zero correspond to blocked values.
    """

    # Default Effelsberg geometry
    pr = 50  # Primary reflector radius
    sr = 3.25  # Sub-reflector radius

    L = 20   # Length support structure (from the edge of the sr)
    a = 1  # Half-width support structure

    # Angle shade effect in aperture plane
    alpha = np.radians(10)  # triangle angle

    block = np.zeros(x.shape)  # or y.shape same
    block[(x ** 2 + y ** 2 < pr ** 2) & (x ** 2 + y ** 2 > sr ** 2)] = 1
    block[(-(sr + L) < x) & (x < (sr + L)) & (-a < y) & (y < a)] = 0
    block[(-(sr + L) < y) & (y < (sr + L)) & (-a < x) & (x < a)] = 0

    csc2 = np.sin(alpha) ** (-2)  # squared cosecant

    # base of the triangle
    d = (-a + np.sqrt(a ** 2 - (a ** 2 - pr ** 2) * csc2)) / csc2

    # points for the triangle coordinates
    A = sr + L
    B = a
    C = d / np.tan(alpha)
    D = a + d

    y1 = line_equation((A, B), (C, D), x)
    y2 = line_equation((A, -B), (C, -D), x)
    x3 = line_equation((-A, B), (-C, D), y)
    x4 = line_equation((-A, -B), (-C, -D), y)
    y5 = line_equation((-A, -B), (-C, -D), x)
    y6 = line_equation((-A, B), (-C, D), x)
    x7 = line_equation((A, -B), (C, -D), y)
    x8 = line_equation((A, B), (C, D), y)

    def circ(s):
        return np.sqrt(np.abs(pr ** 2 - s ** 2))

    block[(A < x) & (C > x) & (y1 > y) & (y2 < y)] = 0
    block[(pr > x) & (C < x) & (circ(x) > y) & (-circ(x) < y)] = 0

    block[(-A > y) & (-C < y) & (x4 < x) & (x3 > x)] = 0
    block[(-pr < y) & (-C > y) & (circ(y) > x) & (-circ(y) < x)] = 0

    block[(-A > x) & (-C < x) & (y5 < y) & (y6 > y)] = 0
    block[(-pr < x) & (-C > x) & (circ(x) > y) & (-circ(x) < y)] = 0

    block[(A < y) & (C > y) & (x7 < x) & (x8 > x)] = 0
    block[(pr > y) & (C < y) & (circ(x) > y) & (-circ(x) < y)] = 0

    return block


def block_manual(pr, sr, a, L):
    """
    Truncation for the aperture (amplitude) distribution, :math:`B(x, y)`,
    manual set up for the primary radius (**pr**), sub-reflector radius
    (**sr**), half-width of a support leg (**a**) and length of the support
    leg (**L**) measured from the edge of the sub-reflector radius. It has
    been considered 4 support legs. To exclude **sr**, **a** or **L** set them
    to zero.

    Parameters
    ----------
    pr : `float`
        Primary reflector radius, in meters.
    sr : `float`
        Sub-reflector radius, in meters.
    a : `float`
        Half-width of a support leg, in meters.
    L : `float`
        Length of a support leg, measured from the edge of the sub-reflector
        towards its end, in meters.

    Returns
    -------
    block_func : `function`
        It returns the function ``block_func(x, y)``, which depends only on the
        grid values, similar to `~pyoof.telgeometry.block_effelsberg`.
    """

    def block_func(x, y):
        block = np.zeros(x.shape)  # or y.shape same
        block[(x ** 2 + y ** 2 < pr ** 2) & (x ** 2 + y ** 2 > sr ** 2)] = 1
        block[(-(sr + L) < x) & (x < (sr + L)) & (-a < y) & (y < a)] = 0
        block[(-(sr + L) < y) & (y < (sr + L)) & (-a < x) & (x < a)] = 0
        return block

    return block_func



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import ascii
from astropy.utils.data import get_pkg_data_filename
from scipy import interpolate, optimize
import os
import time
import yaml
from .aperture import radiation_pattern, phase
from .math_functions import wavevector2radians, co_matrices
from .plot_routines import plot_fit_path
from .aux_functions import store_data_csv, illum_strings, store_data_ascii

__all__ = [
    'residual_true', 'residual', 'params_complete', 'fit_beam',
    ]


def residual_true(
    params, beam_data_norm, u_data, v_data, d_z, wavel, illum_func, telgeo,
    resolution, box_factor, interp
        ):
    """
    Computes the true residual ready to use for the `~pyoof.fit_beam`
    function. True means that some of the parameters used will not be fitted.
    Their selection is done by default or by adding a ``config_params.yml``
    file to the `~pyoof.fit_beam` function.

    Parameters
    ----------
    params : `~numpy.ndarray`
        Two stacked arrays, the illumination and Zernike circle polynomials
        coefficients. ``params = np.hstack([I_coeff, K_coeff])``.
    beam_data_norm : `list`
        The ``beam_data_norm`` is a list with the three observed beam maps,
        :math:`P^\\mathrm{obs}_\\mathrm{norm}(u, v)`, minus, zero and plus
        out-of-focus. The data has to be initially normalized by its maximum.
    u_data : `~numpy.ndarray`
        :math:`x` axis value for the 3 beam maps in radians. The values have
        to be flatten, in one dimension, and stacked in the same order as the
        ``d_z = [d_z-, 0., d_z+]`` values from each beam map.
    v_data : `~numpy.ndarray`
        :math:`y` axis value for the 3 beam maps in radians. The values have
        to be flatten, one dimensional, and stacked in the same order as the
        ``d_z = [d_z-, 0., d_z+]`` values from each beam map.
    d_z : `list`
        Radial offset :math:`d_z`, added to the sub-reflector in meters. This
        characteristic measurement adds the classical interference pattern to
        the beam maps, normalized squared (field) radiation pattern, which is
        an out-of-focus property. The radial offset list must be as follows,
        ``d_z = [d_z-, 0., d_z+]`` all of them in meters.
    wavel : `float`
        Wavelength, :math:`\\lambda`, of the observation in meters.
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with the key **I_coeff**. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.
    telgeo : `list`
        List that contains the blockage distribution, optical path difference
        (OPD) function, and the primary radius (`float`) in meters. The list
        must have the following order, ``telego = [block_dist, opd_func, pr]``.
    resolution : `int`
        Fast Fourier Transform resolution for a rectangular grid. The input
        value has to be greater or equal to the telescope resolution and with
        power of 2 for faster FFT processing. It is recommended a value higher
        than ``resolution = 2 ** 8``.
    box_factor : `int`
        Related to the FFT resolution (**resolution** key), defines the image
        pixel size level. It depends on the primary radius, ``pr``, of the
        telescope, e.g. a ``box_factor = 5`` returns ``x = np.linspace(-5 *
        pr, 5 * pr, resolution)``, an array to be used in the FFT2
        (`~numpy.fft.fft2`).
    interp : `bool`
        If `True`, it will process the correspondent interpolation between
        the observed grid (:math:`P^\\mathrm{obs}_\\mathrm{norm}(u, v)`) and
        the computed grid (:math:`P_\\mathrm{norm}(u, v)`) for the FFT2
        aperture distribution model (:math:`\\underline{E_\\mathrm{a}}(x,
        y)`).

    Returns
    -------
    _residual_true : `~numpy.ndarray`
        One dimensional array of the residual between the observed data and
        the FFT aperture distribution model. It has been concatenated as
        minus, zero and plus radial offset (to do a multiple fit). It is
        required to have the residual in one dimension in order to use a least
        squares minimization `~scipy.optimize.least_squares` package.
    """

    I_coeff = params[:4]
    K_coeff = params[4:]

    beam_model = []
    for i in range(3):

        u, v, F = radiation_pattern(
            K_coeff=K_coeff,
            d_z=d_z[i],
            wavel=wavel,
            I_coeff=I_coeff,
            illum_func=illum_func,
            telgeo=telgeo,
            resolution=resolution,
            box_factor=box_factor
            )

        power_pattern = np.abs(F) ** 2

        # Normalized power pattern model (Beam model)
        power_norm = power_pattern / power_pattern.max()

        if interp:

            # Generated beam u and v: wave-vectors -> degrees -> radians
            u_rad = wavevector2radians(u, wavel)
            v_rad = wavevector2radians(v, wavel)

            # The calculated beam needs to be transformed!
            intrp = interpolate.RegularGridInterpolator(
                points=(u_rad, v_rad),  # points defining grid
                values=power_norm.T,    # data in grid
                method='linear'         # linear or nearest
                )

            # input interpolation function is the real beam grid
            beam_model.append(intrp(np.array([u_data[i], v_data[i]]).T))

        else:
            beam_model.append(power_norm)

    beam_model_all = np.hstack((beam_model[0], beam_model[1], beam_model[2]))
    beam_data_all = np.hstack(
        (beam_data_norm[0], beam_data_norm[1], beam_data_norm[2])
        )

    # Residual = data - model
    _residual_true = beam_data_all - beam_model_all

    return _residual_true


def residual(
    params, idx, N_K_coeff, beam_data_norm, u_data, v_data, d_z, wavel,
    illum_func, telgeo, resolution, box_factor, interp, config_params
        ):
    """
    Wrapper for the `~pyoof.residual_true` function. The objective of
    this function is to fool the `~scipy.optimize.least_squares` package by
    changing the number of parameters that will be used in the fit. The
    parameter array must be organized as follows, ``params = np.hstack([
    I_coeff, K_coeff])``. The parameter selection is done by default or by
    adding a ``config_params.yml`` file to the `~pyoof.fit_beam` function.

    Parameters
    ----------
    params : `~numpy.ndarray`
        Two stacked arrays, the illumination and Zernike circle polynomials
        coefficients. ``params = np.hstack([I_coeff, K_coeff])``.
    idx : `list`
        List of the positions for the removed parameters for the least squares minimization in the ``params`` array.
        on.
    N_K_coeff : `int`
        Total number of Zernike circle polynomials coefficients to fit. It is
        obtained from the order to be fitted with the formula
        ``N_K_coeff = (n + 1) * (n + 2) // 2.``.
    beam_data_norm : `list`
        The ``beam_data_norm`` is a list with the three observed beam maps,
        :math:`P^\\mathrm{obs}_\\mathrm{norm}(u, v)`, minus, zero and plus
        out-of-focus. The data has to be initially normalized by its maximum.
    u_data : `~numpy.ndarray`
        :math:`x` axis value for the 3 beam maps in radians. The values have
        to be flatten, in one dimension, and stacked in the same order as the
        ``d_z = [d_z-, 0., d_z+]`` values from each beam map.
    v_data : `~numpy.ndarray`
        :math:`y` axis value for the 3 beam maps in radians. The values have
        to be flatten, one dimensional, and stacked in the same order as the
        ``d_z = [d_z-, 0., d_z+]`` values from each beam map.
    d_z : `list`
        Radial offset :math:`d_z`, added to the sub-reflector in meters. This
        characteristic measurement adds the classical interference pattern to
        the beam maps, normalized squared (field) radiation pattern, which is
        an out-of-focus property. The radial offset list must be as follows,
        ``d_z = [d_z-, 0., d_z+]`` all of them in meters.
    wavel : `float`
        Wavelength, :math:`\\lambda`, of the observation in meters.
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with the key **I_coeff**. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.
    telgeo : `list`
        List that contains the blockage distribution, optical path difference
        (OPD) function, and the primary radius (`float`) in meters. The list
        must have the following order, ``telego = [block_dist, opd_func, pr]``.
    resolution : `int`
        Fast Fourier Transform resolution for a rectangular grid. The input
        value has to be greater or equal to the telescope resolution and with
        power of 2 for faster FFT processing. It is recommended a value higher
        than ``resolution = 2 ** 8``.
    box_factor : `int`
        Related to the FFT resolution (**resolution** key), defines the image
        pixel size level. It depends on the primary radius, ``pr``, of the
        telescope, e.g. a ``box_factor = 5`` returns ``x = np.linspace(-5 *
        pr, 5 * pr, resolution)``, an array to be used in the FFT2
        (`~numpy.fft.fft2`).
    interp : `bool`
        If `True`, it will process the correspondent interpolation between
        the observed grid (:math:`P^\\mathrm{obs}_\\mathrm{norm}(u, v)`) and
        the computed grid (:math:`P_\\mathrm{norm}(u, v)`) for the FFT2
        aperture distribution model (:math:`\\underline{E_\\mathrm{a}}(x,
        y)`).
    config_params : `dict`
        Contains the values for the fixed parameters (excluded from the least
        squares minimization), by default four parameters are kept fixed,
        ``i_amp``, ``x0``, ``y0`` and ``K(0, 0)``. See the
        ``config_params.yml`` file.

    Returns
    -------
    _residual_true : `~numpy.ndarray`
        Same output from `~pyoof.residual_true`.
        One dimensional array of the residual between the observed data and
        the FFT aperture distribution model. It has been concatenated as
        minus, zero and plus radial offset (to do a multiple fit). It is
        required to have the residual in one dimension in order to use a least
        squares minimization `~scipy.optimize.least_squares` package.

    Notes
    -----
    The **idx** key needs an indices list of the parameters to be removed.
    The structure of the parameters always follows, ``params = np.hstack([
    I_coeff, K_coeff])``, a list with ``idx = [0, 1, 2, 4]`` will remove from
    the least squares minimization, ``[i_amp, taper_dB, x0, y0, K(0, 0)]``.
    """

    # Parameters list for the true fit
    params_res = params_complete(params, idx, N_K_coeff, config_params)

    _residual_true = residual_true(
        params=params_res,
        beam_data_norm=beam_data_norm,
        u_data=u_data,
        v_data=v_data,
        d_z=d_z,
        wavel=wavel,
        resolution=resolution,
        box_factor=box_factor,
        illum_func=illum_func,
        telgeo=telgeo,
        interp=interp,
        )

    return _residual_true


def params_complete(params, idx, N_K_coeff, config_params):
    """
    This function fills the missing parameters not used in the lease squares
    minimization, they are required to compute the correct aperture
    distribution, :math:`\\underline{E_\\mathrm{a}}(x, y)`. By default the
    following parameters are excluded ``i_amp``, ``x0``, ``y0``, ``K(0, 0)``. The
    parameter selection is done by default or by adding a
    ``config_params.yml`` file to the `~pyoof.fit_beam` function.

    Parameters
    ----------
    params : `~numpy.ndarray`
        Contains the incomplete array of parameters, the ``params`` array will
        be updated for the correct number of parameters to be used in the
        `~pyoof.residual_true` function. The array should be of the shape,
        ``params = np.hstack([I_coeff, K_coeff])``, missing or not some of the
        ``idx = [0, 1, 2, 3, 4]`` parameters.
    idx : `list`
        List of the positions for the removed parameters for the least squares
        minimization in the ``params`` array.
    N_K_coeff : `int`
        Total number of Zernike circle polynomials coefficients to fit. It is
        obtained from the order to be fitted with the formula
        ``N_K_coeff = (n + 1) * (n + 2) // 2.``.
    config_params : `dict`
        Contains the values for the fixed parameters (excluded from the least
        squares minimization), by default four parameters are kept fixed,
        ``i_amp``, ``x0``, ``y0`` and ``K(0, 0)``. See the
        ``config_params.yml`` file.

    Returns
    -------
    params_updated : `~numpy.ndarray`
        Complete set of parameters to be used in the `~pyoof.residual_true`
        function.
    """

    # Fixed values for parameters, in case they're excluded, see idx
    [i_amp_f, taper_dB_f, x0_f, y0_f, K_f] = config_params['params_fixed']

    # N_K_coeff number of Zernike circle polynomials coefficients
    if params.size != (4 + N_K_coeff):
        params_updated = params.copy()
        for i in idx:
            if i == 0:
                params_updated = np.insert(params_updated, i, i_amp_f)
                # assigned value for i_amp
            elif i == 1:
                params_updated = np.insert(params_updated, i, taper_dB_f)
                # assigned value for c_dB
            elif i == 2:
                params_updated = np.insert(params_updated, i, x0_f)
                # assigned value for x0
            elif i == 3:
                params_updated = np.insert(params_updated, i, y0_f)
                # assigned value for y0
            else:
                params_updated = np.insert(params_updated, i, K_f)
                # assigned value for any other
    else:
        params_updated = params

    return params_updated


def fit_beam(
    data_info, data_obs, method, order_max, illum_func, telescope, resolution,
    box_factor, fit_previous=True, config_params_file=None, make_plots=True,
    verbose=2
        ):
    """
    Computes the Zernike circle polynomial coefficients, ``K_coeff``, and the
    illumination function coefficients, ``I_coeff``, stores and plots data (
    optional) by using a least squares minimization. The stored data belongs
    to the best fitted power pattern (or beam map). `~pyoof.fit_beam` is the
    core function from the `~pyoof` package.

    Parameters
    ----------
    data_info : `list`
        It contains all extra data besides the beam map. The output
        corresponds to a list,
        ``[name, pthto, obs_object, obs_date, freq, wavel, d_z, meanel]``.
        These are, name of the fits file, paht of the fits file, observed
        object, observation date, frequency, wavelength, radial offset and
        mean elevation, respectively.
    data_obs : `list`
        It contains beam maps and :math:`x`-, and :math:`y`-axis
        (:math:`uv`-plane in Fourier space) data for the least squares
        minimization (see `~pyoof.fit_beam`). The list has the following order
        ``[beam_data, u_data, v_data]``. ``beam_data`` is the three beam
        observations, minus, zero and plus out-of-focus, in a flat array.
        ``u_data`` and ``v_data`` are the beam axes in a flat array.
    method : `str`
        Least squares minimization algorithm, it can be ``'trf'``, ``'lm'`` or
        ``'dogbox'``. ``'lm'`` does not handle bounds, see documentation
        `~scipy.optimize.least_squares`.
    order_max : `int`
        Maximum order used for the Zernike circle polynomials, :math:`n`, least
        squares minimization. If ``order_max = 3``, it will do the
        optimization for orders 1, 2 and 3.
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with the key **I_coeff**. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.
    telescope : `list`
        List that contains the blockage distribution, optical path difference
        (OPD) function, primary radius (`float`) in meters, and telescope name
        (`str`). The list must have the following order, ``telescope =
        [block_dist, opd_func, pr, tel_name]``.
    resolution : `int`
        Fast Fourier Transform resolution for a rectangular grid. The input
        value has to be greater or equal to the telescope resolution and with
        power of 2 for faster FFT processing. It is recommended a value higher
        than ``resolution = 2 ** 8``.
    box_factor : `int`
        Related to the FFT resolution (**resolution** key), defines the image
        pixel size level. It depends on the primary radius, ``pr``, of the
        telescope, e.g. a ``box_factor = 5`` returns ``x = np.linspace(-5 *
        pr, 5 * pr, resolution)``, an array to be used in the FFT2
        (`~numpy.fft.fft2`).
    fit_previous : `bool`
        If set to `True`, it will fit the coefficients from the previous
        optimization this feature is strongly suggested. If `False`, it will
        find the new coefficients by using the standard initial coefficients.
    config_params_file : `str`
        Path for the configuration file, this includes, the maximum and
        minimum bounds, excluded, fixed and initial parameters for the
        optimization. See ``config_params.yml`` in the pyoof package directory.
    make_plots : `bool`
        If `True` will generate a sub-directory with all the important plots
        for the OOF holography, including the phase error, :math:`\\varphi(x,
        y)` and fitted beam, :math:`P_\\mathrm{norm}(u, v)`.
    verbose : `int`
        {0, 1, 2} Level of algorithm verbosity. 0 work silent, 1 display
        termination report, 2, display progress during iteration (default).
    """

    start_time = time.time()

    print('\n ******* PYOOF FIT POWER PATTERN ******* \n')
    print('... Reading data ... \n')

    # All observed data needed to fit the beam
    [name, pthto, obs_object, obs_date, freq, wavel, d_z, meanel] = data_info
    [beam_data, u_data, v_data] = data_obs

    illum_name, taper_name = illum_strings(illum_func)

    try:
        telgeo, tel_name = telescope[:3], telescope[3]

        # Calling default configuration file from the pyoof package
        if config_params_file is None:
            config_params_pyoof = get_pkg_data_filename(
                'data/config_params.yml'
                )

            with open(config_params_pyoof, 'r') as yaml_config:
                config_params = yaml.load(yaml_config)
        else:
            with open(config_params_file, 'r') as yaml_config:
                config_params = yaml.load(yaml_config)

        # Generating specific exceptions
        if not callable(illum_func):
            raise ValueError('illum_func must be a function')

        if not (
            callable(telescope[0]) and callable(telescope[1]) and
            isinstance(telescope[2], float) and isinstance(telescope[3], str)
                ):
            raise ValueError(
                'telescope has to be a list [func, func, float, str]'
                )

    except ValueError as error:
        print(error.args)
    except NameError:
        print(
            'Configuration file .yml does not exist in path: ' +
            config_params_file
            )

    else:
        pass

    # Storing files in pyoof_out directory
    # pthto: path or directory where the fits file is located
    if not os.path.exists(pthto + '/pyoof_out'):
        os.makedirs(pthto + '/pyoof_out')

    for j in ["%03d" % i for i in range(101)]:
        name_dir = pthto + '/pyoof_out/' + name + '-' + str(j)
        if not os.path.exists(name_dir):
            os.makedirs(name_dir)
            break

    print('Maximum order to be fitted: ', order_max)
    print('Telescope name: ', tel_name)
    print('File name: ', name)
    print('Obs frequency: ', freq, 'Hz')
    print('Obs Wavelength : ', np.round(wavel, 4), 'm')
    print('d_z (out-of-focus): ', np.round(d_z, 4), 'm')
    print('Illumination to be fitted: ', illum_name)

    for order in range(1, order_max + 1):

        if not verbose == 0:
            print('\n... Fit order ' + str(order) + ' ... \n')

        # Setting limits for plotting fitted beam
        plim_u = [np.min(u_data[0]), np.max(u_data[0])]  # radians
        plim_v = [np.min(v_data[0]), np.max(v_data[0])]  # radians
        plim_rad = np.array(plim_u + plim_v)

        # Beam normalization
        beam_data_norm = [beam_data[i] / beam_data[i].max() for i in range(3)]

        n = order                           # order polynomial to fit
        N_K_coeff = (n + 1) * (n + 2) // 2  # number of K(n, l) to fit

        # Looking for result parameters lower order
        if fit_previous and n != 1:
            N_K_coeff_previous = n * (n + 1) // 2
            path_params_previous = name_dir + '/fitpar_n' + str(n - 1) + '.csv'
            params_to_add = N_K_coeff - N_K_coeff_previous

            if os.path.exists(path_params_previous):
                params_init = np.hstack(
                    (ascii.read(path_params_previous)['parfit'],
                        np.ones(params_to_add) * 0.1)
                    )
                if not verbose == 0:
                    print('Initial params: n={} fit'.format(n - 1))
            else:
                print(
                    '\n ERROR: There is no previous parameters file fitpar_n' +
                    str(n - 1) + '.csv in directory \n'
                    )
        else:
            params_init = np.array(
                config_params['params_init'] + [0.1] * (N_K_coeff - 1)
                )
            print('Initial params: default')
            # i_amp, sigma_r, x0, y0, K(n, l)
            # Giving an initial value of 0.1 for each coeff

        idx = config_params['params_excluded']  # exclude params from fit
        # [0, 1, 2, 3, 4] = [i_amp, c_dB, x0, y0, K(0, 0)]
        # or 'None' to include all
        params_init_true = np.delete(params_init, idx)

        if method == 'lm':  # see scipy.optimize.least_squares
            bounds = tuple([
                -np.ones(params_init_true.shape) * np.inf,
                np.ones(params_init_true.shape) * np.inf
                ])

        else:
            bounds_min = np.array(
                config_params['params_bounds_min'] + [-5] * (N_K_coeff - 1)
                )
            bounds_max = np.array(
                config_params['params_bounds_max'] + [5] * (N_K_coeff - 1)
                )

            bounds_min_true = np.delete(bounds_min, idx)
            bounds_max_true = np.delete(bounds_max, idx)

            bounds = tuple([bounds_min_true, bounds_max_true])

        if not verbose == 0:
            print('Parameters to fit: {}\n'.format(len(params_init_true)))

        # Running nonlinear least squares minimization
        res_lsq = optimize.least_squares(
            fun=residual,
            x0=params_init_true,
            args=(               # Conserve this order in arguments!
                idx,             # Index of parameters to be excluded (params)
                N_K_coeff,       # Total Zernike circle polynomial coeff
                beam_data_norm,  # Normalized beam maps
                u_data,          # x coordinate beam map
                v_data,          # y coordinate beam map
                d_z,             # Radial offset
                wavel,           # Wavelength of observation
                illum_func,      # Illumination function
                telgeo,          # telgeo = [block_dist, opd_func, pr]
                resolution,      # FFT2 resolution for a rectangular grid
                box_factor,      # Image pixel size level
                True,            # Grid interpolation
                config_params    # Coeff configuration for minimization (dict)
                ),
            bounds=bounds,
            method=method,
            verbose=verbose,
            max_nfev=None
            )

        # Solutions from least squared optimization
        params_solution = params_complete(
            params=res_lsq.x,
            idx=idx,
            N_K_coeff=N_K_coeff,
            config_params=config_params
            )
        params_init = params_init               # Initial parameters used
        res_optim = res_lsq.fun.reshape(3, -1)  # Optimum residual solution
        jac_optim = res_lsq.jac                 # Last Jacobian matrix
        grad_optim = res_lsq.grad               # Last Gradient

        cov, corr = co_matrices(
            res=res_lsq.fun,
            jac=res_lsq.jac,
            n_pars=params_init_true.size  # number of parameters fitted
            )
        cov_ptrue = np.vstack((np.delete(np.arange(N_K_coeff + 4), idx), cov))
        corr_ptrue = np.vstack(
            (np.delete(np.arange(N_K_coeff + 4), idx), corr)
            )

        # Final phase from fit in the telescope's primary reflector
        _phase = phase(
            K_coeff=params_solution[4:],
            notilt=True,
            pr=telgeo[2]
            )[2]

        # Storing files in directory
        if not verbose == 0:
            print('... Saving data ... \n')

        store_data_ascii(
            name=name,
            name_dir=name_dir,
            taper_name=taper_name,
            order=n,
            params_solution=params_solution,
            params_init=params_init,
            )

        # Printing the results from saved ascii file
        if not verbose == 0:
            print(ascii.read(name_dir + '/fitpar_n' + str(n) + '.csv'))

        if n == 1:
            pyoof_info = dict(
                telescope=tel_name,
                name=name,
                obs_object=obs_object,
                obs_date=obs_date,
                d_z=d_z,
                wavel=wavel,
                frequency=freq,
                illumination=illum_name,
                meanel=meanel,
                fft_resolution=resolution,
                box_factor=box_factor,
                opt_method=method
                )

            with open(name_dir + '/pyoof_info.yml', 'w') as outfile:
                outfile.write('# pyoof relevant information\n')
                yaml.dump(pyoof_info, outfile, default_flow_style=False)

        # To store large files in csv format
        save_to_csv = [
            beam_data, u_data, v_data, res_optim, jac_optim, grad_optim,
            _phase, cov_ptrue, corr_ptrue
            ]

        store_data_csv(
            name=name,
            name_dir=name_dir,
            order=n,
            save_to_csv=save_to_csv
            )

        if make_plots:
            if not verbose == 0:
                print('\n... Making plots ...')

            plot_fit_path(  # Making all relevant plots
                path_pyoof=name_dir,
                order=n,
                telgeo=telgeo,
                illum_func=illum_func,
                plim_rad=plim_rad,
                save=True,
                angle='degrees',
                resolution=resolution,
                box_factor=box_factor
                )

            plt.close('all')

    final_time = np.round((time.time() - start_time) / 60, 2)
    print('\n **** PYOOF FIT COMPLETED AT {} mins **** \n'.format(final_time))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Tomas Cassanelli
import os
from astropy.io import fits
import numpy as np
from astropy.io import ascii
from scipy.constants import c as light_speed
from .aperture import illum_gauss, illum_pedestal

__all__ = [
    'extract_data_pyoof', 'extract_data_effelsberg', 'str2LaTeX',
    'store_data_csv', 'uv_ratio', 'illum_strings'
    ]


def illum_strings(illum_func):
    """
    It assigns string labels to the illumination function. The `~pyoof` package
    has two standard illumination functions, `~pyoof.aperture.illum_pedestal`
    and `~pyoof.aperture.illum_gauss`.

    Parameters
    ----------
    illum_func : `function`
        Illumination function, :math:`E_\\mathrm{a}(x, y)`, to be evaluated
        with ``I_coeff``. The illumination functions available are
        `~pyoof.aperture.illum_pedestal` and `~pyoof.aperture.illum_gauss`.

    Returns
    -------
    illum_name : `str`
        String with the illumination function name.
    taper_name : `str`
        String with the illumination function taper.
    """

    # adding illumination function information
    if illum_func == illum_pedestal:
        illum_name = 'pedestal'
        taper_name = 'c_dB'
    elif illum_func == illum_gauss:
        illum_name = 'gauss'
        taper_name = 'sigma_dB'
    else:
        illum_name = 'manual'
        taper_name = 'taper_dB'

    return illum_name, taper_name


def extract_data_pyoof(pathfits):
    """
    Extracts data from the `~pyoof` default fits file OOF holography
    observations, ready to use for the least squares minimization (see
    `~pyoof.fit_beam`). The fits file has to have the following keys on its
    PrimaryHDU header: ``'FREQ'``, ``'WAVEL'``, ``'MEANEL'``, ``'OBJECT'`` and
    ``'DATE_OBS'``. Besides this three BinTableHDU are required for the data
    itself; ``MINUS OOF``, ``ZERO OOF`` and ``PLUS OOF``. The BinTableHDU
    header has to have the ``'DZ'`` key which includes the radial offset,
    :math:`d_z`. Finally the BinTableHDU has the data files ``'U'``, ``'V'``
    and ``'BEAM'``, which is the :math:`x`- and :math:`y`-axis position in
    radians and the ``'BEAM'`` in a flat array, in mJy.

    Parameters
    ----------
    pathfits : `str`
        Path to the fits file that contains the three beam maps pre-calibrated,
        using the correct PrimaryHDU and the three BinTableHDU (``MINUS OOF``,
        ``ZERO OOF`` and ``PLUS OOF``).

    Returns
    -------
    data_info : `list`
        It contains all extra data besides the beam map. The output
        corresponds to a list,
        ``[name, pthto, obs_object, obs_date, freq, wavel, d_z, meanel]``.
        These are, name of the fits file, paht of the fits file, observed
        object, observation date, frequency, wavelength, radial offset and
        mean elevation, respectively.
    data_obs : `list`
        It contains beam maps and :math:`x`-, and :math:`y`-axis
        (:math:`uv`-plane in Fourier space) data for the least squares
        minimization (see `~pyoof.fit_beam`). The list has the following order
        ``[beam_data, u_data, v_data]``. ``beam_data`` is the three beam
        observations, minus, zero and plus out-of-focus, in a flat array.
        ``u_data`` and ``v_data`` are the beam axes in a flat array.
    """

    hdulist = fits.open(pathfits)  # open fits file, pyoof format
    # path or directory where the fits file is located
    pthto = os.path.split(pathfits)[0]
    # name of the fit file to fit
    name = os.path.split(pathfits)[1][:-5]

    if not all(
            k in hdulist[0].header
            for k in ['FREQ', 'WAVEL', 'MEANEL', 'OBJECT', 'DATE_OBS']
            ):
        raise TypeError('Not all necessary keys found in FITS header.')

    freq = hdulist[0].header['FREQ']
    wavel = hdulist[0].header['WAVEL']
    meanel = hdulist[0].header['MEANEL']
    obs_object = hdulist[0].header['OBJECT']
    obs_date = hdulist[0].header['DATE_OBS']

    beam_data = [hdulist[i].data['BEAM'] for i in range(1, 4)]
    u_data = [hdulist[i].data['U'] for i in range(1, 4)]
    v_data = [hdulist[i].data['V'] for i in range(1, 4)]
    d_z = [hdulist[i].header['DZ'] for i in range(1, 4)]

    data_file = [name, pthto]
    data_info = data_file + [obs_object, obs_date, freq, wavel, d_z, meanel]
    data_obs = [beam_data, u_data, v_data]

    return data_info, data_obs


def extract_data_effelsberg(pathfits):
    """
    Extracts data from the Effelsberg OOF holography observations, ready to
    use for the least squares minimization. This function will only work for
    the Effelsberg telescope beam maps.

    Parameters
    ----------
    pathfits : `str`
        Path to the fits file that contains the three beam maps pre-calibrated,
        from the Effelsberg telescope.

    Returns
    -------
    data_info : `list`
        It contains all extra data besides the beam map. The output
        corresponds to a list,
        ``[name, pthto, obs_object, obs_date, freq, wavel, d_z, meanel]``.
        These are, name of the fits file, paht of the fits file, observed
        object, observation date, frequency, wavelength, radial offset and
        mean elevation, respectively.
    data_obs : `list`
        It contains beam maps and :math:`x`-, and :math:`y`-axis
        (:math:`uv`-plane in Fourier space) data for the least squares
        minimization (see `~pyoof.fit_beam`). The list has the following order
        ``[beam_data, u_data, v_data]``. ``beam_data`` is the three beam
        observations, minus, zero and plus out-of-focus, in a flat array.
        ``u_data`` and ``v_data`` are the beam axes in a flat array.
    """

    # Opening fits file with astropy
    try:
        # main fits file with the OOF holography format
        hdulist = fits.open(pathfits)

        # Observation frequency
        freq = hdulist[0].header['FREQ']  # Hz
        wavel = light_speed / freq

        # Mean elevation
        meanel = hdulist[0].header['MEANEL']  # Degrees
        obs_object = hdulist[0].header['OBJECT']  # observed object
        obs_date = hdulist[0].header['DATE_OBS']  # observation date
        d_z = [hdulist[i].header['DZ'] for i in range(1, 4)][::-1]

        beam_data = [hdulist[i].data['fnu'] for i in range(1, 4)][::-1]
        u_data = [hdulist[i].data['DX'] for i in range(1, 4)][::-1]
        v_data = [hdulist[i].data['DY'] for i in range(1, 4)][::-1]

    except FileNotFoundError:
        print('Fits file does not exists in directory: ' + pathfits)
    except NameError:
        print('Fits file does not have the OOF holography format')

    else:
        pass

    # Permuting the position to provide same as main_functions
    beam_data.insert(1, beam_data.pop(2))
    u_data.insert(1, u_data.pop(2))
    v_data.insert(1, v_data.pop(2))
    d_z.insert(1, d_z.pop(2))

    # path or directory where the fits file is located
    pthto = os.path.split(pathfits)[0]
    # name of the fit file to fit
    name = os.path.split(pathfits)[1][:-5]

    data_info = [name, pthto, obs_object, obs_date, freq, wavel, d_z, meanel]
    data_obs = [beam_data, u_data, v_data]

    return data_info, data_obs


def str2LaTeX(python_string):
    """
    Function that solves the underscore problem in a python string to
    :math:`\LaTeX` string.

    Parameters
    ----------
    python_string : `str`
        String that needs to be changed.

    Returns
    -------
    LaTeX_string : `str`
        String with the new underscore symbol.
    """

    string_list = list(python_string)
    for idx, string in enumerate(string_list):
        if string_list[idx] == '_':
            string_list[idx] = '\\_'

    LaTeX_string = ''.join(string_list)

    return LaTeX_string


def store_data_csv(name, name_dir, order, save_to_csv):
    """
    Stores all important information in a csv file after the least squares
    minimization has finished, `~pyoof.fit_beam`. All data will be stores in
    the ``pyoof_out/name`` directory, with ``name`` the name of the fits file.

    Parameters
    ----------
    name : `str`
        File name of the fits file to be optimized.
    name_dir : `str`
        Path to store all the csv files. The files will depend on the order of
        the Zernike circle polynomial.
    order : `int`
        Order used for the Zernike circle polynomial, :math:`n`.
    save_to_csv : `list`
        It contains all data that will be stored. The list must have the
        following order, ``[beam_data, u_data, v_data, res_optim, jac_optim,
        grad_optim, phase, cov_ptrue, corr_ptrue]``.
    """

    headers = [
        'Normalized beam', 'u vector radians', 'v vector radians', 'Residual',
        'Jacobian', 'Gradient', 'Phase primary reflector radians',
        'Variance-Covariance matrix (first row fitted parameters idx)',
        'Correlation matrix (first row fitted parameters idx)'
        ]

    fnames = [
        '/beam_data.csv', '/u_data.csv', '/v_data.csv',
        '/res_n' + str(order) + '.csv', '/jac_n' + str(order) + '.csv',
        '/grad_n' + str(order) + '.csv', '/phase_n' + str(order) + '.csv',
        '/cov_n' + str(order) + '.csv', '/corr_n' + str(order) + '.csv'
        ]

    if order != 1:
        headers = headers[3:]
        fnames = fnames[3:]
        save_to_csv = save_to_csv[3:]

    for fname, header, file in zip(fnames, headers, save_to_csv):
        np.savetxt(
            fname=name_dir + fname,
            X=file,
            header=header + ' ' + name
            )


def store_data_ascii(
    name, name_dir, taper_name, order, params_solution, params_init
        ):
    """
    Stores in an ascii format the parameters found by the least squares
    minimization (see `~pyoof.fit_beam`).

    Parameters
    ----------
    name : `str`
        File name of the fits file to be optimized.
    name_dir : `str`
        Path to store all the csv files. The files will depend on the order of
        the Zernike circle polynomial.
    taper_name : `str`
        Name of the illumination function taper.
    order : `int`
        Order used for the Zernike circle polynomial, :math:`n`.
    params_solution : `~numpy.ndarray`
        Contains the best fitted parameters, the illumination function
        coefficients, ``I_coeff`` and the Zernike circle polynomial
        coefficients, ``K_coeff`` in one array.
    params_init : `~numpt.ndarray`
        Contains the initial parameters used in the least squares minimization
        to start finding the best fitted combination of them.
    """

    n = order
    N_K_coeff = (n + 1) * (n + 2) // 2

    # Making nice table :)
    ln = [(j, i) for i in range(0, n + 1) for j in range(-i, i + 1, 2)]
    L = np.array(ln)[:, 0]
    N = np.array(ln)[:, 1]

    params_names = ['i_amp', taper_name, 'x_0', 'y_0']
    for i in range(N_K_coeff):
        params_names.append('K({}, {})'.format(N[i], L[i]))

    # To store fit information and found parameters in ascii file
    ascii.write(
        table=[params_names, params_solution, params_init],
        output=name_dir + '/fitpar_n' + str(n) + '.csv',
        names=['parname', 'parfit', 'parinit'],
        comment='Fitted parameters ' + name
        )


def uv_ratio(u, v):
    """
    Calculates the aspect ratio for the 3 power pattern plots, plus some
    corrections for the text on it. Used in the `function` `~pyoof.plot_beam`
    and `~pyoof.plot_data`

    Parameters
    ----------
    u : `~numpy.ndarray`
        Spatial frequencies from the power pattern, usually in degrees.
    v : `~numpy.ndarray`
        Spatial frequencies from the power pattern, usually in degrees.

    Returns
    -------
    plot_width : `float`
        Width for the power pattern figure.
    plot_height : `float`
        Height for the power pattern figure.
    """

    ratio = (v.max() - v.min()) / (3 * (u.max() - u.min()))
    width = 14
    height = width * (ratio) + 0.2

    return width, height
